/*  ===========================================================================
*
*   This file is part of HISE.
*   Copyright 2016 Christoph Hart
*
*   HISE is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option any later version.
*
*   HISE is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with HISE.  If not, see <http://www.gnu.org/licenses/>.
*
*   Commercial licences for using HISE in an closed source project are
*   available on request. Please visit the project's website to get more
*   information about commercial licencing:
*
*   http://www.hartinstruments.net/hise/
*
*   HISE is based on the JUCE library,
*   which also must be licenced for commercial applications:
*
*   http://www.juce.com
*
*   ===========================================================================
*/


namespace snex {
namespace jit {
using namespace juce;
USE_ASMJIT_NAMESPACE;



int Compiler::compileCount = 0;

 void Compiler::reset()
 {
	 if (compiler != nullptr)
		 delete compiler;

	 handler = new NamespaceHandler();
	 compiler = new ClassCompiler(&memory, *handler);
	 memory.registerFunctionsToNamespaceHandler(getNamespaceHandler());
 }

 bool Compiler::allowInlining() const
 {
	 return memory.getOptimizationPassList().contains(OptimizationIds::Inlining);
 }

 snex::jit::FunctionClass::Map Compiler::getFunctionMap()
 {
	 auto m = memory.getMap();

	 if (auto ifc = getInbuiltFunctionClass())
		 m.addArray(ifc->getMap());
	 
	 for (auto t : handler->getComplexTypeList())
	 {
		 FunctionClass::Ptr fc = t->getFunctionClass();
		 auto fm = fc->getMap();

		 m.addArray(fm);
	 }

	 return m;
 }


Compiler::Compiler(GlobalScope& memoryPool):
	memory(memoryPool),
	cr(Result::ok())
{
	reset();
}

Compiler::~Compiler()
{
	if(compiler != nullptr)
		delete compiler;
}


juce::String Compiler::getAssemblyCode()
{
	return assembly;
}



juce::ValueTree Compiler::getAST() const
{
	if (compiler->syntaxTree != nullptr)
	{
		return compiler->syntaxTree->toValueTree();
	}

	return {};
}

juce::String Compiler::dumpSyntaxTree() const
{
	if (compiler->syntaxTree != nullptr)
	{
		return dynamic_cast<SyntaxTree*>(compiler->syntaxTree.get())->dump();
	}

	return {};
}

juce::String Compiler::dumpNamespaceTree() const
{
	return compiler->namespaceHandler.dump();
}

ComplexType::Ptr Compiler::registerExternalComplexType(ComplexType::Ptr t)
{
	if (auto st = dynamic_cast<StructType*>(t.get()))
	{
		st->setCompiler(*this);
		st->finaliseExternalDefinition();
	}

	return compiler->namespaceHandler.registerComplexTypeOrReturnExisting(t);
}

ComplexType::Ptr Compiler::getComplexType(const NamespacedIdentifier& s, const Array<TemplateParameter>& tp, bool checkPreexisting)
{
	if (tp.isEmpty())
	{
		return compiler->namespaceHandler.getComplexType(s);
	}
	else
	{
		if (checkPreexisting)
		{
			if (auto c = compiler->namespaceHandler.getExistingTemplateInstantiation(s, tp))
				return c;
		}

		auto r = Result::ok();
		return compiler->namespaceHandler.createTemplateInstantiation({ s, {} }, tp, r);
	}
}

void Compiler::addConstant(const NamespacedIdentifier& s, const VariableStorage& v)
{
    NamespaceHandler::SymbolDebugInfo di;
	compiler->namespaceHandler.addSymbol(s, TypeInfo(v.getType(), true, false), NamespaceHandler::Constant, di);
	compiler->namespaceHandler.addConstant(s, v);
}

void Compiler::addTemplateClass(const TemplateObject& c)
{
	compiler->namespaceHandler.addTemplateClass(c);
}

snex::jit::NamespaceHandler& Compiler::getNamespaceHandler()
{
	return compiler->namespaceHandler;
}

snex::jit::FunctionClass::Ptr Compiler::getInbuiltFunctionClass()
{
	return compiler->getInbuiltFunctionClass();
}

void Compiler::initInbuildFunctions()
{
	compiler->setInbuildFunctions();
}

juce::Result Compiler::getCompileResult()
{
	return cr;
}


JitObject Compiler::compileJitObject(const juce::String& code)
{
	compileCount++;
	lastCode = code;
	
	try
	{
		Preprocessor p(lastCode);
		p.addDefinitionsFromScope(memory.getPreprocessorDefinitions());
		preprocessedCode = p.process();
	}
	catch (ParserHelpers::Error& e)
	{
		compiler->lastResult = Result::fail(e.toString());
	}
	catch (juce::String& e)
	{
		compiler->lastResult = Result::fail(e);
		return {};
	}
	

	

	JitObject snexObject(compiler->compileAndGetScope(preprocessedCode));

	cr = compiler->getLastResult();

#if SNEX_MIR_BACKEND

	if (cr.wasOk())
	{
		auto layout = compiler->namespaceHandler.createDataLayouts();
		
		mir::MirCompiler mc(memory);

		mc.setDataLayout(layout);

		JitObject mirObject(mc.compileMirCode(getAST()));

		cr = mc.getLastError();

#if SNEX_INCLUDE_NMD_ASSEMBLY

		assembly = {};

		for (const auto& fid : mirObject.getFunctionIds())
		{
			auto f = mirObject[fid.toString()];

			assembly << "\t;" << f.getSignature({}, false) << "\n";

			assembly << f.createAssembly() << "\n";
		}
#else
		assembly = mc.getAssembly();
#endif

		return mirObject;
	}
	else
	{
		return {};
	}

	
#else
	assembly = compiler->assembly;
	return snexObject;
#endif
}


snex::jit::NamespaceHandler& Compiler::parseWithoutCompilation(const juce::String& code)
{
	try
	{
		Preprocessor p(code);
		p.addDefinitionsFromScope(memory.getPreprocessorDefinitions());
		preprocessedCode = p.process();
	}
	catch (juce::String& e)
	{
		compiler->lastResult = Result::fail(e);
		return compiler->namespaceHandler;
	}

	compiler->parseOnly = true;
	compiler->compileAndGetScope(preprocessedCode);

	return compiler->namespaceHandler;
}

void Compiler::setDebugHandler(DebugHandler* newHandler, bool useLineNumbersInErrorMessage)
{
	compiler->setDebugHandler(newHandler);
	compiler->parentScope->getGlobalScope()->addDebugHandler(newHandler);
	compiler->setUseCodeInErrorMessage(!useLineNumbersInErrorMessage);
}

struct SyntaxTreeDictionaryProvider : public zstd::DictionaryProviderBase<void>
{
public:

	SyntaxTreeDictionaryProvider(InputStream* /*input*/ = nullptr) :
		zstd::DictionaryProviderBase<void>(nullptr)
	{}

	MemoryBlock createDictionaryData() override
	{
		static const unsigned char dictionary[] = { 55, 164, 48, 236, 200, 129, 119, 40, 77, 16, 136, 51, 213, 1, 0, 0, 0, 0, 0, 0, 0, 0, 8, 0, 0, 0, 64, 145, 170, 20, 1, 48, 144, 251, 161, 28, 25, 21, 76, 176, 140, 57, 50, 44, 245, 203, 66, 106, 92, 180, 199, 195, 123, 60, 230, 56, 247, 134, 153, 51, 193,
			206, 148, 254, 95, 146, 23, 69, 166, 90, 221, 39, 109, 155, 54, 105, 59, 201, 179, 66, 213, 230, 82, 190, 252, 1, 243, 0, 0, 0, 6, 134, 198, 38, 180, 49, 149, 188, 142, 0, 4, 224, 192, 130, 132, 6, 11, 141, 137, 133, 133, 5, 133, 7, 7, 133, 133, 4, 3, 6, 131,
			132, 4, 131, 133, 6, 132, 4, 67, 97, 33, 89, 156, 32, 148, 204, 150, 72, 103, 157, 147, 3, 0, 0, 132, 152, 138, 200, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 4, 0, 0, 0, 8, 0, 0, 0, 116, 70, 114, 111, 109, 0, 1,
			6, 73, 110, 108, 105, 5, 118, 111, 105, 100, 0, 0, 67, 111, 109, 112, 121, 109, 98, 111, 108, 0, 1, 55, 5, 105, 110, 100, 101, 120, 58, 111, 112, 101, 73, 100, 0, 1, 33, 5, 105, 110, 100, 101, 120, 58, 48, 44, 32, 48, 62, 62, 0, 73, 100, 115, 0, 1, 12, 5,
			109, 97, 105, 110, 58, 58, 105, 100, 120, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 27, 5, 123, 32, 123, 32, 48, 46, 102, 32, 125, 44, 32, 111, 112, 101, 100, 58, 58, 108, 101, 110, 103, 116, 104, 0, 0, 86, 97, 114, 105, 97, 1, 3, 5, 45, 0,
			1, 2, 68, 111, 116, 0, 1, 3, 5, 60, 0, 1, 2, 68, 111, 116, 0, 102, 108, 111, 97, 116, 44, 32, 53, 62, 32, 109, 97, 105, 110, 58, 101, 58, 58, 108, 111, 111, 112, 69, 110, 100, 0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110,
			101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 53, 0, 0, 73, 97, 109, 101, 0, 1, 42, 5, 105, 110, 116, 0, 0, 0, 0, 1, 4, 65, 110, 111, 110, 121, 109, 111, 97, 116, 101, 100, 58, 58, 120, 48, 0, 1, 1, 86, 97, 114, 105,
			108, 97, 116, 101, 100, 58, 58, 120, 49, 0, 1, 1, 86, 97, 114, 105, 108, 0, 1, 30, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 105, 48, 0, 0, 86, 97, 114, 105, 101, 116, 65, 108, 112,
			104, 97, 0, 1, 4, 73, 110, 108, 105, 115, 116, 97, 114, 116, 0, 0, 86, 97, 114, 105, 97, 80, 97, 114, 97, 109, 49, 41, 0, 67, 116, 104, 105, 115, 0, 1, 1, 68, 111, 116, 0, 73, 110, 100, 101, 120, 58, 58, 108, 105, 109, 105, 116, 0, 0, 68, 111, 116, 0, 1,
			99, 97, 108, 101, 100, 0, 0, 66, 105, 110, 97, 114, 99, 97, 108, 101, 100, 0, 0, 68, 111, 116, 0, 1, 116, 97, 0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1,
			3, 5, 48, 0, 0, 73, 110, 108, 105, 110, 102, 108, 111, 97, 116, 44, 32, 51, 62, 0, 73, 100, 115, 0, 1, 13, 5, 109, 97, 105, 110, 58, 58, 100, 97, 116, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 19, 5, 123, 32, 49, 46, 102, 44, 121,
			109, 98, 111, 108, 0, 1, 46, 5, 105, 110, 116, 32, 105, 110, 112, 83, 116, 97, 114, 116, 0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 49, 0, 0, 73,
			110, 108, 105, 110, 111, 108, 0, 1, 30, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 58, 58, 118, 97, 108, 100, 60, 48, 44, 32, 48, 62, 62, 0, 0, 65, 115, 115, 105, 121, 109, 98, 111, 108, 0, 1, 68, 5, 105,
			110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 60, 105, 110, 100, 101, 120, 58, 58, 102, 32, 125, 32, 125, 0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 5, 5, 51,
			46, 102, 0, 0, 73, 110, 108, 105, 115, 116, 97, 114, 116, 0, 0, 84, 101, 114, 110, 97, 114, 121, 67, 97, 108, 108, 0, 1, 2, 66, 105, 110, 97, 116, 70, 114, 111, 109, 58, 58, 115, 99, 97, 108, 101, 100, 0, 0, 65, 115, 115, 105, 1, 29, 5, 105, 110, 116, 32, 105,
			110, 100, 101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 58, 58, 108, 101, 110, 103, 116, 104, 0, 0, 73, 110, 108, 105, 110, 98, 111, 108, 0, 1, 28, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 58, 58, 115, 116, 97,
			114, 116, 0, 0, 65, 115, 115, 105, 99, 97, 108, 101, 100, 58, 58, 115, 101, 116, 76, 111, 111, 112, 82, 97, 110, 103, 101, 0, 1, 6, 73, 110, 108, 105, 110, 101, 100, 99, 97, 108, 101, 100, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 0, 1, 4, 73, 110, 0, 84, 121,
			112, 101, 0, 1, 45, 5, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 60, 102, 108, 111, 97, 116, 44, 32, 105, 110, 100, 101, 120, 58, 121, 112, 101, 0, 1, 4, 5, 33, 61, 0, 1, 2, 68, 111, 116, 0, 1, 3, 5, 61, 0, 1, 2, 84, 101,
			114, 110, 97, 114, 121, 115, 0, 1, 19, 5, 123, 32, 48, 46, 102, 44, 32, 49, 46, 102, 44, 32, 50, 46, 102, 32, 125, 0, 0, 82, 112, 69, 110, 100, 0, 0, 73, 110, 108, 105, 110, 101, 100, 80, 97, 114, 97, 109, 101, 116, 101, 114, 0, 1, 2, 76, 105, 110, 101, 0,
			1, 5, 1, 0, 0, 0, 0, 83, 121, 109, 98, 111, 108, 0, 1, 44, 5, 105, 110, 116, 97, 109, 101, 0, 1, 44, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 108, 111, 111, 112, 101, 100, 58, 58, 115, 101, 116, 76, 111, 111, 112, 82, 97, 110, 103, 101, 58, 58,
			108, 111, 111, 112, 83, 116, 97, 114, 116, 0, 111, 111, 112, 101, 100, 58, 58, 105, 110, 105, 116, 86, 97, 108, 117, 101, 0, 0, 73, 97, 108, 117, 101, 115, 0, 1, 25, 5, 123, 32, 123, 32, 48, 32, 125, 44, 32, 123, 32, 48, 32, 125, 44, 32, 123, 32, 48, 32, 125, 32,
			58, 108, 101, 110, 103, 116, 104, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 48, 0, 0, 68, 111, 116, 0, 1, 111, 111, 112, 101, 100, 58, 58, 115, 116, 97,
			114, 116, 0, 0, 70, 117, 110, 99, 1, 35, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 108, 111, 111, 112, 101, 100, 58, 58, 103, 101, 58, 108, 111, 111, 112, 101, 100, 60, 48, 44, 32, 48, 62, 0, 0, 65, 110, 100, 101, 120, 58, 58, 108, 111, 111, 112, 101, 100,
			58, 58, 108, 111, 111, 112, 101, 100, 0, 1, 4, 73, 98, 111, 108, 0, 1, 37, 5, 102, 108, 111, 97, 116, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 50, 49, 58, 58, 115, 0, 0, 86, 97, 114, 0,
			1, 9, 6, 0, 192, 5, 10, 1, 0, 0, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 3, 76, 111, 111, 112, 0, 1, 4, 111, 97, 116, 44, 32, 56, 62, 32, 100, 97, 116, 97, 0,
			0, 83, 98, 111, 108, 0, 1, 15, 5, 102, 108, 111, 97, 116, 32, 109, 97, 105, 110, 58, 58, 120, 0, 0, 82, 101, 116, 117, 114, 114, 97, 116, 111, 114, 0, 1, 38, 5, 102, 108, 111, 97, 116, 38, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115,
			83, 99, 111, 112, 101, 76, 105, 110, 101, 50, 48, 58, 58, 115, 0, 1, 2, 115, 0, 1, 44, 5, 123, 32, 49, 46, 102, 44, 32, 50, 46, 102, 44, 32, 51, 46, 102, 44, 32, 52, 46, 102, 44, 32, 53, 46, 102, 44, 32, 54, 46, 102, 44, 32, 55, 46, 102, 44, 32, 56,
			46, 102, 32, 125, 0, 0, 70, 111, 108, 0, 1, 11, 5, 88, 32, 109, 97, 105, 110, 58, 58, 120, 0, 0, 86, 97, 114, 105, 0, 84, 121, 112, 101, 0, 1, 13, 5, 115, 112, 97, 110, 60, 88, 44, 32, 49, 50, 62, 0, 73, 100, 115, 0, 97, 108, 117, 101, 115, 0, 1,
			31, 5, 123, 32, 123, 32, 50, 46, 102, 32, 125, 44, 32, 123, 32, 51, 46, 102, 32, 125, 44, 32, 123, 32, 48, 46, 102, 32, 125, 32, 125, 0, 0, 82, 73, 100, 115, 0, 1, 4, 5, 118, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 19, 5, 123, 32,
			56, 46, 102, 44, 32, 48, 46, 102, 44, 32, 49, 46, 102, 32, 125, 0, 105, 110, 58, 58, 118, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 19, 5, 123, 32, 50, 46, 102, 44, 32, 51, 46, 102, 44, 32, 52, 46, 102, 32, 125, 0, 0, 82, 108, 117, 101,
			0, 1, 5, 5, 48, 46, 102, 0, 0, 86, 97, 114, 98, 111, 108, 0, 1, 17, 5, 84, 119, 111, 73, 110, 116, 115, 32, 109, 97, 105, 110, 58, 58, 99, 0, 0, 73, 109, 109, 101, 100, 42, 5, 118, 111, 105, 100, 32, 99, 108, 101, 97, 114, 83, 112, 97, 110, 40, 84, 119,
			111, 73, 110, 116, 115, 38, 32, 99, 108, 101, 97, 114, 83, 112, 97, 110, 58, 58, 100, 97, 116, 97, 41, 0, 67, 97, 108, 108, 84, 98, 111, 108, 0, 1, 18, 5, 100, 111, 117, 98, 108, 101, 32, 79, 98, 106, 101, 99, 116, 58, 58, 100, 0, 0, 83, 121, 110, 116, 97, 121,
			112, 101, 0, 1, 8, 5, 79, 98, 106, 101, 99, 116, 0, 73, 100, 115, 0, 1, 20, 5, 109, 97, 105, 110, 58, 58, 100, 49, 44, 109, 97, 105, 110, 58, 58, 100, 50, 44, 0, 73, 110, 105, 116, 0, 1, 3, 5, 51, 0, 0, 86, 97, 114, 105, 97, 98, 111, 108, 0, 1,
			22, 5, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32, 51, 62, 32, 109, 97, 105, 110, 58, 58, 100, 0, 0, 73, 110, 108, 105, 97, 109, 101, 0, 1, 39, 5, 105, 110, 116, 32, 125, 32, 125, 0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110,
			101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 52, 0, 0, 73, 110, 108, 105, 110, 97, 108, 117, 101, 115, 0, 1, 14, 5, 123, 32, 50, 46, 102, 44, 32, 51, 46, 102, 32, 125, 0, 0, 82, 101, 116, 114, 0, 1, 9, 6, 0, 16, 154,
			9, 1, 0, 0, 0, 1, 1, 6, 5, 88, 58, 58, 89, 0, 1, 1, 83, 121, 110, 116, 116, 84, 121, 112, 101, 0, 1, 12, 5, 115, 112, 97, 110, 60, 88, 44, 32, 50, 62, 0, 1, 110, 60, 88, 44, 32, 50, 62, 0, 73, 100, 115, 0, 1, 7, 5, 100, 97, 116, 97,
			44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 15, 5, 123, 32, 123, 32, 123, 32, 57, 32, 125, 32, 125, 32, 125, 0, 5, 88, 58, 58, 73, 110, 110, 101, 114, 0, 73, 100, 115, 0, 1, 6, 5, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117,
			101, 115, 0, 1, 7, 5, 123, 32, 50, 32, 125, 0, 0, 70, 117, 110, 98, 111, 108, 0, 1, 24, 5, 105, 110, 116, 32, 88, 58, 58, 73, 110, 110, 101, 114, 58, 58, 73, 110, 110, 101, 114, 58, 58, 122, 0, 0, 68, 111, 116, 0, 1, 73, 100, 115, 0, 1, 7, 5, 121,
			79, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 12, 5, 123, 32, 123, 32, 57, 48, 32, 1, 15, 5, 105, 110, 116, 32, 89, 58, 58, 121, 86, 97, 108, 117, 101, 0, 0, 67, 98, 111, 108, 0, 1, 12, 5, 89, 38, 32, 88, 58, 58, 88, 58,
			58, 111, 0, 0, 86, 97, 114, 105, 97, 101, 114, 0, 1, 9, 6, 64, 192, 5, 10, 1, 0, 0, 0, 98, 111, 108, 0, 1, 19, 5, 78, 101, 115, 116, 101, 100, 32, 109, 97, 105, 110, 58, 58, 111, 98, 106, 50, 0, 0, 83, 121, 110, 116, 97, 116, 0, 73, 100, 115, 0,
			1, 13, 5, 78, 101, 115, 116, 101, 100, 58, 58, 111, 50, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 12, 5, 123, 32, 123, 32, 52, 48, 32, 125, 32, 125, 0, 0, 98, 111, 108, 0, 1, 21, 5, 77, 121, 79, 98, 106, 101, 99, 116, 32, 78, 101, 115,
			116, 101, 100, 58, 58, 111, 50, 0, 0, 86, 97, 114, 105, 121, 112, 101, 0, 1, 8, 5, 78, 101, 115, 116, 101, 100, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 4, 70, 117, 110, 99,
			116, 105, 111, 1, 17, 5, 105, 110, 116, 32, 77, 121, 79, 98, 106, 101, 99, 116, 58, 58, 118, 0, 0, 66, 105, 110, 97, 114, 98, 111, 108, 0, 1, 19, 5, 77, 121, 79, 98, 106, 101, 99, 116, 32, 109, 97, 105, 110, 58, 58, 111, 49, 0, 0, 66, 105, 110, 97, 121, 112,
			101, 0, 1, 10, 5, 77, 121, 79, 98, 106, 101, 99, 116, 0, 73, 100, 115, 0, 1, 11, 5, 109, 97, 105, 110, 58, 58, 111, 51, 44, 0, 73, 110, 105, 116, 86, 97, 116, 117, 114, 101, 0, 1, 40, 5, 118, 111, 105, 100, 32, 77, 121, 79, 98, 106, 101, 99, 116, 58, 58,
			77, 121, 79, 98, 106, 101, 99, 116, 40, 105, 110, 116, 32, 105, 110, 105, 116, 86, 97, 108, 117, 101, 41, 0, 44, 32, 105, 62, 0, 73, 100, 115, 0, 1, 6, 5, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 29, 5, 123, 32, 123, 32, 123,
			32, 123, 32, 57, 48, 32, 125, 32, 125, 32, 125, 44, 32, 123, 117, 114, 101, 0, 1, 30, 5, 118, 111, 105, 100, 32, 105, 58, 58, 105, 40, 119, 114, 97, 112, 58, 58, 101, 118, 101, 110, 116, 60, 101, 62, 38, 32, 111, 41, 0, 1, 1, 83, 121, 110, 98, 111, 108, 0, 1,
			14, 5, 88, 32, 109, 97, 105, 110, 58, 58, 111, 98, 106, 50, 0, 0, 82, 101, 116, 117, 114, 97, 108, 117, 101, 115, 0, 1, 18, 5, 123, 32, 123, 32, 49, 32, 125, 44, 32, 123, 32, 52, 32, 125, 32, 125, 0, 0, 70, 101, 114, 0, 1, 9, 6, 128, 192, 5, 10, 1,
			0, 0, 0, 1, 1, 83, 98, 111, 108, 0, 1, 14, 5, 87, 114, 97, 112, 112, 101, 114, 60, 88, 62, 32, 119, 0, 0, 86, 97, 114, 105, 97, 88, 0, 73, 100, 115, 0, 1, 15, 5, 87, 114, 97, 112, 112, 101, 114, 58, 58, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86,
			97, 108, 117, 101, 115, 0, 1, 12, 5, 123, 32, 123, 32, 48, 46, 32, 125, 32, 125, 0, 0, 67, 97, 116, 117, 114, 101, 0, 1, 42, 5, 118, 111, 105, 100, 32, 87, 114, 97, 112, 112, 101, 114, 58, 58, 115, 101, 116, 80, 97, 114, 97, 109, 101, 116, 101, 114, 60, 50, 48,
			62, 40, 100, 111, 117, 98, 111, 108, 0, 1, 11, 5, 105, 110, 116, 32, 88, 50, 58, 58, 118, 0, 0, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 4, 67, 108, 97, 115, 116, 84, 121, 112, 101, 0,
			1, 20, 5, 119, 114, 97, 112, 58, 58, 105, 110, 105, 116, 60, 88, 50, 44, 32, 73, 50, 62, 0, 77, 101, 109, 98, 101, 114, 84, 121, 112, 101, 0, 1, 5, 5, 88, 50, 38, 0, 79, 102, 102, 115, 73, 100, 115, 0, 1, 4, 5, 111, 44, 0, 73, 110, 105, 116, 86, 97,
			108, 117, 101, 115, 0, 1, 24, 5, 123, 32, 123, 32, 123, 32, 56, 32, 125, 32, 125, 44, 32, 123, 32, 111, 98, 106, 32, 125, 32, 125, 0, 0, 67, 1, 43, 5, 118, 111, 105, 100, 32, 88, 50, 58, 58, 115, 101, 116, 80, 97, 114, 97, 109, 101, 116, 101, 114, 60, 48, 62,
			40, 100, 111, 117, 98, 108, 101, 32, 110, 101, 119, 86, 97, 108, 117, 101, 41, 0, 1, 1, 83, 121, 110, 0, 1, 3, 5, 48, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0,
			1, 3, 5, 48, 0, 0, 65, 115, 115, 105, 103, 98, 111, 108, 0, 1, 19, 5, 88, 32, 119, 114, 97, 112, 58, 58, 105, 110, 105, 116, 58, 58, 111, 98, 106, 0, 0, 86, 97, 0, 84, 121, 112, 101, 0, 1, 18, 5, 119, 114, 97, 112, 58, 58, 105, 110, 105, 116, 60, 88,
			44, 32, 73, 62, 0, 73, 100, 115, 0, 1, 108, 117, 101, 0, 1, 3, 5, 88, 0, 0, 84, 101, 109, 112, 108, 97, 116, 101, 100, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 0, 70, 117, 110, 99, 116, 121, 112,
			101, 0, 1, 3, 5, 89, 0, 73, 100, 115, 0, 1, 10, 5, 109, 97, 105, 110, 58, 58, 121, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 6, 5, 123, 32, 32, 125, 0, 1, 1, 86, 97, 114, 105, 117, 114, 101, 0, 1, 17, 5, 118, 111, 105, 100, 32,
			89, 58, 58, 89, 40, 88, 38, 32, 120, 41, 0, 1, 1, 83, 121, 110, 118, 97, 108, 117, 101, 0, 0, 66, 105, 110, 97, 114, 121, 79, 112, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3, 5, 43, 0, 1, 2,
			68, 111, 116, 0, 101, 0, 1, 3, 5, 88, 0, 77, 101, 109, 98, 101, 114, 84, 121, 112, 101, 0, 1, 6, 5, 88, 58, 58, 90, 0, 79, 102, 102, 115, 101, 116, 0, 1, 5, 1, 16, 0, 0, 0, 1, 1, 86, 97, 123, 32, 49, 48, 48, 48, 32, 125, 32, 125, 44, 32,
			123, 32, 123, 32, 48, 32, 125, 123, 32, 49, 48, 48, 48, 32, 125, 32, 125, 44, 32, 123, 32, 123, 32, 48, 46, 32, 125, 32, 125, 32, 125, 0, 0, 65, 110, 111, 110, 121, 1, 17, 5, 105, 110, 116, 32, 88, 58, 58, 89, 58, 58, 118, 97, 108, 117, 101, 0, 0, 82, 101,
			58, 88, 40, 105, 110, 116, 32, 88, 58, 58, 88, 58, 58, 118, 41, 0, 67, 97, 108, 108, 84, 121, 112, 101, 0, 1, 16, 5, 77, 101, 109, 98, 101, 114, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 2, 86, 97, 114, 1, 7, 5, 105, 110, 116, 32, 118, 0, 0, 70, 117,
			110, 99, 117, 114, 101, 0, 1, 18, 5, 118, 111, 105, 100, 32, 88, 58, 58, 88, 40, 105, 110, 116, 32, 118, 41, 0, 1, 1, 83, 88, 0, 73, 100, 115, 0, 1, 23, 5, 109, 97, 105, 110, 58, 58, 111, 98, 106, 44, 109, 97, 105, 110, 58, 58, 111, 98, 106, 50, 44, 97,
			108, 117, 101, 115, 0, 1, 15, 5, 123, 32, 123, 32, 123, 32, 48, 32, 125, 32, 125, 32, 125, 0, 97, 116, 117, 114, 101, 0, 1, 16, 5, 118, 111, 105, 100, 32, 88, 58, 58, 89, 58, 58, 89, 40, 41, 0, 1, 1, 83, 121, 97, 108, 117, 101, 0, 0, 67, 111, 109, 112,
			108, 118, 97, 108, 117, 101, 0, 0, 68, 111, 116, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 83, 117, 98, 115, 101, 0, 1, 3, 5, 88, 0, 0, 86, 97, 114, 105, 97, 114, 0, 1, 9, 6, 64, 160, 224, 9, 1, 0, 0, 0, 1, 108,
			0, 1, 40, 5, 68, 101, 114, 105, 118, 101, 100, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 111, 98, 106, 0, 0, 70, 117, 110, 99, 116, 101, 114, 0, 1, 9, 6, 0, 161, 224, 9, 1,
			0, 0, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 83, 116, 97, 116, 101, 109, 101, 97, 116, 117, 114, 101, 0, 1, 19, 5, 118, 111, 105, 100, 32, 66, 97, 115, 101, 58, 58, 66,
			97, 115, 101, 40, 41, 0, 1, 1, 83, 121, 121, 109, 98, 111, 108, 0, 1, 27, 5, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 49, 62, 32, 109, 97, 105, 110, 58, 58, 100, 97, 116, 97, 0, 0, 82, 101, 116, 117, 114, 97, 116, 44, 32, 49, 62, 0, 73, 100,
			115, 0, 1, 13, 5, 109, 97, 105, 110, 58, 58, 100, 97, 116, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 9, 5, 123, 32, 48, 46, 102, 32, 125, 0, 0, 70, 97, 116, 117, 114, 101, 0, 1, 45, 5, 118, 111, 105, 100, 32, 100, 99, 58, 58, 112,
			114, 111, 99, 101, 115, 115, 70, 114, 97, 109, 101, 60, 49, 62, 40, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 49, 62, 38, 32, 97, 116, 117, 114, 101, 0, 1, 36, 5, 118, 111, 105, 100, 32, 100, 99, 58, 58, 115, 101, 116, 80, 97, 114, 97, 109, 101, 116, 101,
			114, 60, 48, 62, 40, 100, 111, 117, 98, 108, 101, 32, 118, 41, 0, 1, 1, 83, 97, 108, 117, 101, 115, 0, 1, 22, 5, 123, 32, 123, 32, 123, 32, 48, 32, 125, 32, 125, 44, 32, 123, 32, 49, 32, 125, 32, 125, 0, 0, 70, 0, 73, 68, 0, 1, 8, 5, 78, 111, 100,
			101, 73, 100, 0, 86, 97, 108, 117, 101, 0, 1, 4, 5, 100, 99, 0, 0, 84, 101, 109, 112, 108, 97, 116, 0, 84, 121, 112, 101, 0, 1, 9, 5, 68, 101, 114, 105, 118, 101, 100, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101,
			0, 1, 5, 1, 0, 0, 0, 0, 1, 1, 65, 115, 115, 105, 98, 111, 108, 0, 1, 16, 5, 105, 110, 116, 32, 68, 101, 114, 105, 118, 101, 100, 58, 58, 100, 0, 0, 67, 111, 109, 112, 108, 97, 116, 117, 114, 101, 0, 1, 27, 5, 105, 110, 116, 32, 68, 101, 114, 105, 118,
			101, 100, 58, 58, 103, 101, 116, 68, 111, 117, 98, 108, 101, 88, 40, 41, 0, 1, 1, 83, 0, 1, 4, 67, 108, 97, 115, 115, 83, 116, 97, 116, 101, 109, 101, 110, 116, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 84, 121, 112, 101, 0, 1, 6, 5,
			66, 97, 115, 101, 0, 1, 1, 83, 121, 110, 1, 13, 5, 105, 110, 116, 32, 66, 97, 115, 101, 58, 58, 121, 0, 0, 67, 108, 97, 115, 115, 0, 1, 13, 5, 68, 101, 114, 105, 118, 101, 100, 32, 111, 98, 106, 0, 0, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1,
			1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 5, 67, 108, 97, 115, 115, 83, 100, 0, 73, 100, 115, 0, 1, 6, 5, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 27, 5, 123, 32, 123, 32, 57, 32, 125, 44, 32, 123, 32,
			49, 48, 32, 125, 44, 32, 123, 32, 57, 112, 101, 0, 1, 9, 5, 68, 101, 114, 105, 118, 101, 100, 0, 73, 100, 115, 0, 1, 12, 5, 109, 97, 105, 110, 58, 58, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 19, 5, 123, 32, 123, 32, 1,
			21, 5, 105, 110, 116, 32, 68, 101, 114, 105, 118, 101, 100, 58, 58, 103, 101, 116, 88, 40, 41, 0, 1, 1, 83, 0, 1, 3, 5, 53, 0, 0, 86, 97, 114, 105, 97, 98, 111, 108, 0, 1, 32, 5, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 60, 53,
			44, 32, 48, 62, 32, 109, 97, 105, 110, 58, 58, 105, 100, 120, 0, 0, 114, 97, 112, 112, 101, 100, 60, 53, 44, 32, 48, 62, 0, 73, 100, 115, 0, 1, 12, 5, 109, 97, 105, 110, 58, 58, 105, 100, 120, 44, 0, 73, 110, 1, 34, 5, 105, 110, 116, 32, 105, 110, 100, 101,
			120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 105, 100, 120, 0, 0, 86, 97, 114, 105, 97, 118, 111, 105, 100, 0, 0, 65, 115, 115, 105, 103, 109, 97, 105, 110, 58, 58, 115, 101, 99, 111, 110, 100, 73, 110, 100, 101, 120, 0,
			0, 73, 110, 108, 111, 112, 101, 73, 100, 0, 1, 25, 5, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58, 119, 114, 97, 112, 112, 101, 100, 0, 1, 4, 73, 110, 1, 3, 5, 49, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76,
			105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 4, 5, 49, 48, 0, 0, 86, 97, 114, 105, 20, 5, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32, 49, 48, 62, 32, 100, 97, 116, 97, 0, 101, 120, 0, 0, 73, 110, 108, 105, 110, 101, 100,
			70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 83, 99, 111, 112, 101, 73, 100, 0, 1, 32, 5, 105, 110, 100, 101, 120, 114, 97, 112, 112, 101, 100, 60, 49, 48, 44, 32, 48, 62, 0, 73, 100, 115, 0, 1, 19, 5,
			109, 97, 105, 110, 58, 58, 102, 105, 114, 115, 116, 73, 110, 100, 101, 120, 44, 0, 73, 110, 105, 116, 97, 109, 101, 0, 1, 38, 5, 117, 110, 114, 101, 115, 111, 108, 118, 101, 100, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 103, 101, 116, 70, 114,
			111, 109, 58, 58, 99, 0, 1, 1, 86, 97, 114, 105, 110, 60, 105, 110, 116, 44, 32, 51, 50, 62, 0, 73, 100, 115, 0, 1, 7, 5, 100, 97, 116, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 7, 5, 123, 32, 57, 32, 125, 0, 0, 70, 117, 110,
			98, 111, 108, 0, 1, 42, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58, 43, 43, 111, 112, 101, 114, 97, 116, 111, 114, 58, 58, 110, 101, 119, 86, 97, 108, 117, 101, 0, 0, 68, 121, 109, 98, 111, 108, 0, 1, 25, 5, 105,
			110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 60, 51, 50, 44, 32, 48, 62, 32, 106, 0, 0, 65, 115, 115, 105, 121, 112, 101, 0, 1, 23, 5, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 60, 51, 50, 44, 32, 48, 62, 0, 73, 100, 115,
			0, 1, 4, 5, 106, 44, 0, 73, 110, 105, 116, 1, 3, 5, 61, 0, 1, 2, 86, 97, 114, 0, 84, 121, 112, 101, 0, 1, 22, 5, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 60, 48, 44, 32, 48, 62, 38, 0, 1, 1, 80, 111, 114, 78, 97, 109, 101,
			0, 1, 33, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 111, 112, 101, 114, 97, 116, 111, 114, 61, 58, 58, 118, 0, 1, 1, 73, 109, 109, 101, 67, 97, 108, 108, 0, 1, 2, 86, 97, 114, 105, 97, 98, 108, 101, 82, 101, 102,
			101, 114, 101, 110, 99, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 83, 121, 109, 98, 111, 108, 0, 1, 41, 5, 105, 98, 111, 108, 0, 1, 18, 5, 78, 111, 114, 109, 97, 108, 105, 115, 101, 100, 84, 121, 112, 101, 32, 105, 0, 0, 65, 115, 115,
			114, 78, 97, 109, 101, 0, 1, 44, 5, 102, 108, 111, 0, 84, 121, 112, 101, 0, 1, 16, 5, 78, 111, 114, 109, 97, 108, 105, 115, 101, 100, 84, 121, 112, 101, 0, 73, 100, 115, 0, 1, 4, 5, 118, 111, 105, 100, 0, 0, 82, 101, 116, 117, 114, 116, 104, 105, 115, 0, 1,
			1, 86, 97, 114, 105, 121, 109, 98, 111, 108, 0, 1, 37, 5, 105, 110, 116, 32, 105, 110, 5, 118, 111, 105, 100, 0, 0, 73, 110, 108, 105, 0, 0, 0, 0, 1, 2, 65, 110, 111, 110, 121, 109, 111, 98, 111, 108, 0, 1, 19, 5, 73, 110, 100, 101, 120, 84, 121, 112, 101,
			32, 109, 97, 105, 110, 58, 58, 106, 0, 0, 65, 115, 115, 105, 103, 111, 112, 101, 73, 100, 0, 1, 26, 5, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 116, 121, 112, 101, 95, 110, 115, 97, 102, 101, 58, 58, 105, 110, 105, 116, 86, 97, 108, 117, 101,
			0, 1, 1, 86, 97, 114, 105, 84, 121, 112, 101, 0, 73, 100, 115, 0, 1, 10, 5, 109, 97, 105, 110, 58, 58, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 11, 5, 123, 32, 123, 32, 48, 32, 125, 32, 125, 0, 1, 1, 86, 97, 114, 111, 112, 101,
			73, 100, 0, 1, 23, 5, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 117, 110, 115, 97, 102, 101, 0, 1, 4, 73, 110, 108, 105, 98, 111, 108, 0, 1, 24, 5, 70, 108, 111, 97, 116, 73, 110, 100, 101, 120, 84, 121, 112, 101, 32, 109, 97, 105, 110,
			58, 58, 105, 0, 0, 73, 110, 108, 105, 0, 1, 3, 5, 49, 0, 0, 86, 97, 114, 105, 97, 98, 108, 101, 82, 101, 102, 101, 114, 101, 110, 99, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 83, 121, 109, 98, 111, 108, 0, 1, 50, 5, 102, 97,
			116, 117, 114, 101, 0, 1, 68, 5, 102, 108, 111, 97, 116, 32, 77, 97, 116, 104, 58, 58, 114, 97, 110, 103, 101, 40, 102, 108, 111, 97, 116, 32, 111, 112, 101, 73, 100, 0, 1, 31, 5, 105, 110, 100, 101, 120, 58, 58, 110, 111, 114, 109, 97, 108, 105, 115, 101, 100, 58, 58,
			110, 111, 114, 109, 97, 108, 105, 115, 101, 100, 0, 1, 4, 73, 110, 108, 105, 117, 101, 0, 1, 4, 5, 49, 55, 0, 0, 86, 97, 114, 105, 97, 101, 73, 100, 0, 1, 28, 5, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58, 45, 45, 111, 112, 101,
			114, 97, 116, 111, 114, 0, 1, 4, 73, 110, 108, 105, 112, 101, 95, 99, 97, 115, 116, 0, 1, 3, 73, 110, 108, 105, 110, 101, 100, 97, 108, 58, 58, 118, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0,
			86, 97, 108, 117, 101, 0, 1, 4, 5, 49, 50, 0, 0, 68, 111, 116, 0, 1, 98, 111, 108, 0, 1, 13, 5, 73, 110, 100, 101, 120, 84, 121, 112, 101, 32, 105, 0, 0, 73, 110, 108, 116, 101, 100, 58, 58, 120, 48, 0, 0, 66, 105, 110, 97, 114, 121, 79, 112, 0, 1,
			2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3, 5, 42, 0, 1, 2, 66, 105, 110, 97, 111, 112, 101, 73, 100, 0, 1, 30, 5, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 73, 110, 116, 101,
			114, 112, 111, 108, 97, 116, 101, 100, 0, 1, 5, 73, 110, 108, 105, 121, 109, 98, 111, 108, 0, 1, 42, 5, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 50, 62, 38, 32, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 70, 114, 111, 109,
			58, 58, 99, 49, 0, 108, 0, 1, 40, 5, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 50, 62, 32, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 100, 0, 0, 83, 116, 97, 116, 101, 97, 116, 44, 32, 50,
			62, 38, 0, 73, 100, 115, 0, 1, 27, 5, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 99, 48, 44, 0, 1, 1, 83, 117, 98, 115, 99, 114, 105, 0, 1, 3, 5, 51, 0, 0, 73, 110, 108, 105, 110, 5, 105, 110,
			116, 0, 1, 1, 86, 97, 114, 105, 97, 98, 108, 101, 82, 101, 102, 101, 114, 101, 110, 99, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 83, 121, 109, 98, 111, 108, 0, 1, 39, 5, 105, 111, 108, 0, 1, 39, 5, 105, 110, 116, 32, 105, 110, 100,
			101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 58, 58, 103, 101, 116, 73, 110, 100, 101, 120, 58, 58, 115, 99, 97, 108, 101, 100, 0, 0, 86, 97, 114, 105, 97, 32, 116, 104, 105, 115, 0, 0, 86, 97, 114, 105, 97, 98, 108, 101, 82, 101, 102, 101, 114, 101, 110, 99,
			101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 83, 121, 109, 98, 111, 108, 0, 1, 30, 5, 102, 114, 78, 97, 109, 101, 0, 1, 38, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 58, 58, 103, 101, 116,
			73, 110, 100, 101, 120, 58, 58, 100, 101, 108, 116, 97, 0, 1, 1, 73, 109, 0, 84, 121, 112, 101, 0, 1, 20, 5, 73, 110, 116, 101, 114, 112, 111, 108, 97, 116, 105, 111, 110, 84, 121, 112, 101, 38, 0, 1, 1, 80, 111, 105, 114, 78, 97, 109, 101, 0, 1, 42, 5, 102,
			108, 111, 97, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 99, 97, 108, 101, 100, 58, 58, 97, 115, 115, 125, 44, 32, 123, 32, 50, 46, 102, 49, 57, 46, 102, 32, 125, 44, 32, 123, 32, 51, 46, 53, 102, 53, 46, 102, 32, 125, 32, 125, 0, 0, 70, 121, 112, 101,
			0, 1, 25, 5, 115, 112, 97, 110, 60, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 50, 62, 44, 32, 51, 62, 0, 73, 100, 115, 0, 65, 114, 103, 0, 1, 5, 1, 1, 0, 0, 0, 80, 97, 114, 97, 109, 101, 116, 101, 114, 78, 97, 109, 101, 0, 1, 40, 5,
			102, 108, 111, 97, 116, 32, 98, 111, 108, 0, 1, 35, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 97, 108, 112, 104, 97, 0, 0, 65, 115, 115, 105, 103, 108, 111, 97, 116, 0, 1, 1,
			66, 105, 110, 97, 114, 121, 79, 112, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3, 5, 45, 0, 1, 2, 86, 97, 114, 105, 3, 5, 61, 0, 1, 2, 73, 110, 108, 105, 110, 101, 100, 58, 108, 105, 109, 105, 116,
			0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 50, 0, 0, 73, 110, 108, 105, 110, 98, 111, 108, 0, 1, 40, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120,
			58, 58, 110, 111, 114, 109, 97, 108, 105, 115, 101, 100, 58, 58, 103, 101, 116, 73, 110, 100, 101, 120, 58, 58, 100, 101, 108, 116, 97, 0, 0, 86, 97, 114, 1, 3, 5, 61, 0, 1, 2, 66, 105, 110, 97, 114, 121, 79, 112, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1,
			0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3, 5, 43, 0, 1, 2, 67, 97, 115, 116, 121, 109, 98, 111, 108, 0, 1, 21, 5, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 50, 62, 32, 100, 97, 116, 97, 0, 0, 73, 110, 108, 105, 110, 32, 116, 104,
			105, 115, 0, 1, 1, 84, 104, 105, 115, 80, 111, 105, 110, 116, 101, 114, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 84, 121, 112, 101, 0, 1, 11, 5, 70, 108, 111, 97, 116, 84, 108, 0, 1, 33, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101,
			120, 58, 58, 108, 101, 114, 112, 58, 58, 111, 112, 101, 114, 97, 116, 111, 114, 61, 58, 58, 118, 0, 0, 73, 110, 108, 105, 110, 108, 0, 1, 26, 5, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 60, 70, 108, 111, 97, 116, 84, 121, 112, 101, 62, 32, 105, 0, 0, 73,
			110, 108, 105, 110, 0, 84, 121, 112, 101, 0, 1, 24, 5, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 60, 70, 108, 111, 97, 116, 84, 121, 112, 101, 62, 0, 0, 73, 110, 108, 105, 110, 78, 97, 109, 101, 0, 1, 43, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101,
			120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 73, 110, 116, 101, 114, 112, 111, 108, 97, 116, 101, 100, 58, 58, 97, 108, 112, 104, 97, 0, 98, 111, 108, 0, 1, 36, 5, 117, 110, 114, 101, 115, 111, 108, 118, 101, 100, 32, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114,
			112, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 99, 0, 0, 86, 97, 114, 105, 97, 98, 111, 108, 0, 1, 38, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101, 120, 58, 58, 110, 111, 114, 109, 97, 108, 105, 115, 101, 100, 58, 58, 103, 101, 116, 65, 108, 112, 104, 97,
			58, 58, 102, 0, 0, 73, 110, 108, 105, 110, 116, 104, 58, 58, 114, 97, 110, 103, 101, 40, 105, 110, 116, 32, 105, 110, 112, 117, 116, 44, 32, 105, 110, 116, 32, 114, 97, 110, 103, 101, 58, 58, 80, 97, 114, 97, 109, 49, 44, 32, 105, 110, 116, 32, 114, 121, 109, 98, 111, 108,
			0, 1, 32, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101, 120, 58, 58, 110, 111, 114, 109, 97, 108, 105, 115, 101, 100, 58, 58, 118, 97, 108, 117, 101, 0, 0, 67, 97, 115, 116, 0, 108, 105, 115, 101, 100, 58, 58, 103, 101, 116, 73, 110, 100, 101, 120, 58, 58, 115, 99,
			97, 108, 101, 100, 0, 0, 73, 110, 108, 105, 110, 98, 111, 108, 0, 1, 28, 5, 70, 108, 111, 97, 116, 84, 121, 112, 101, 32, 105, 110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 105, 100, 120, 0, 0, 65, 115, 115, 105, 111, 112, 101, 73, 100, 0, 1, 22, 5, 105,
			110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 103, 101, 116, 70, 114, 111, 109, 0, 1, 8, 73, 110, 108, 105, 110, 101, 100, 108, 105, 115, 101, 100, 58, 58, 97, 115, 115, 105, 103, 110, 73, 110, 116, 101, 114, 110, 97, 108, 58, 58, 118, 0, 0, 68, 111, 116, 0, 1,
			114, 78, 97, 109, 101, 0, 1, 39, 5, 102, 108, 111, 97, 116, 32, 105, 110, 100, 101, 120, 58, 58, 110, 111, 114, 109, 97, 108, 105, 115, 101, 100, 58, 58, 111, 112, 101, 114, 97, 116, 111, 114, 61, 58, 58, 118, 0, 1, 1, 73, 111, 112, 101, 73, 100, 0, 1, 29, 5, 105,
			110, 100, 101, 120, 58, 58, 108, 101, 114, 112, 58, 58, 97, 115, 115, 105, 103, 110, 73, 110, 116, 101, 114, 110, 97, 108, 0, 1, 4, 73, 110, 108, 105, 98, 111, 108, 0, 1, 38, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58,
			116, 121, 112, 101, 95, 99, 97, 115, 116, 58, 58, 108, 105, 109, 105, 116, 0, 0, 73, 110, 108, 105, 110, 0, 84, 121, 112, 101, 0, 1, 11, 5, 73, 110, 100, 101, 120, 84, 121, 112, 101, 0, 0, 65, 115, 115, 105, 103, 97, 112, 112, 101, 100, 58, 58, 119, 114, 97, 112, 112,
			101, 100, 58, 58, 105, 110, 105, 116, 86, 97, 108, 117, 101, 0, 1, 1, 73, 109, 109, 101, 117, 114, 101, 0, 1, 26, 5, 105, 110, 116, 32, 67, 111, 117, 110, 116, 101, 114, 58, 58, 116, 121, 112, 101, 95, 99, 97, 115, 116, 40, 41, 0, 67, 97, 108, 117, 114, 101, 0, 1,
			37, 5, 67, 111, 117, 110, 116, 101, 114, 60, 49, 50, 53, 62, 38, 32, 67, 111, 117, 110, 116, 101, 114, 58, 58, 43, 43, 111, 112, 101, 114, 97, 116, 111, 114, 40, 41, 0, 1, 1, 83, 121, 0, 84, 121, 112, 101, 0, 1, 25, 5, 67, 111, 117, 110, 116, 101, 114, 60, 105,
			110, 116, 32, 85, 112, 112, 101, 114, 76, 105, 109, 105, 116, 62, 0, 1, 1, 67, 108, 97, 115, 115, 83, 116, 1, 4, 5, 49, 50, 0, 0, 86, 97, 114, 105, 97, 98, 108, 101, 82, 101, 102, 101, 114, 101, 110, 99, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0,
			0, 0, 0, 83, 121, 109, 98, 111, 108, 0, 1, 11, 5, 105, 101, 114, 0, 1, 9, 6, 0, 176, 223, 9, 1, 0, 0, 0, 1, 1, 97, 116, 117, 114, 101, 0, 1, 48, 5, 105, 110, 116, 32, 104, 97, 110, 100, 108, 101, 77, 111, 100, 117, 108, 97, 116, 105, 111, 110, 40,
			105, 110, 116, 38, 32, 104, 97, 110, 100, 108, 101, 97, 116, 117, 114, 101, 0, 1, 20, 5, 118, 111, 105, 100, 32, 100, 111, 83, 111, 109, 101, 116, 104, 105, 110, 103, 40, 41, 0, 70, 117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 0, 1, 9, 6, 64, 144, 221, 9, 1, 0,
			0, 1, 3, 5, 43, 0, 1, 2, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 5, 5, 49, 46, 102, 0, 0, 86, 97, 114, 105, 97, 121, 109, 98, 111, 108, 0, 1, 22, 5,
			102, 108, 111, 97, 116, 32, 111, 119, 110, 114, 97, 109, 112, 58, 58, 100, 101, 108, 116, 97, 0, 0, 70, 117, 110, 99, 116, 112, 101, 67, 97, 108, 108, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86,
			97, 108, 117, 101, 0, 1, 4, 5, 50, 46, 0, 0, 82, 101, 116, 117, 114, 108, 117, 101, 0, 1, 3, 5, 56, 0, 0, 82, 101, 116, 117, 114, 117, 101, 0, 1, 4, 5, 49, 50, 0, 0, 83, 116, 97, 116, 101, 108, 117, 101, 0, 1, 5, 5, 57, 48, 48, 0, 0, 86,
			97, 114, 105, 101, 114, 0, 1, 9, 6, 0, 192, 62, 12, 1, 0, 0, 0, 1, 108, 117, 101, 0, 1, 4, 5, 45, 49, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1,
			3, 5, 48, 0, 0, 76, 111, 111, 0, 1, 3, 5, 49, 0, 0, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 3, 67, 111, 109, 112, 108, 101, 120, 97, 116, 117, 114, 101, 0, 1, 85, 5, 98, 108,
			111, 99, 107, 32, 80, 114, 111, 99, 101, 115, 115, 68, 97, 116, 97, 58, 58, 116, 111, 67, 104, 97, 110, 110, 101, 108, 68, 97, 116, 97, 40, 67, 104, 97, 110, 110, 101, 108, 80, 97, 116, 117, 114, 101, 0, 1, 37, 5, 102, 108, 111, 97, 116, 32, 99, 111, 110, 100, 71, 101,
			116, 40, 102, 108, 111, 97, 116, 32, 99, 111, 110, 100, 71, 101, 116, 58, 58, 105, 110, 112, 117, 116, 41, 0, 121, 109, 98, 111, 108, 0, 1, 18, 5, 105, 110, 116, 32, 111, 116, 104, 101, 114, 58, 58, 105, 110, 112, 117, 116, 0, 0, 73, 109, 109, 101, 100, 110, 116, 32, 88,
			58, 58, 118, 97, 108, 117, 101, 0, 0, 83, 121, 110, 116, 97, 98, 111, 108, 0, 1, 7, 5, 88, 32, 111, 98, 106, 0, 0, 86, 97, 114, 105, 97, 88, 0, 73, 100, 115, 0, 1, 6, 5, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 12,
			5, 123, 32, 123, 32, 49, 50, 32, 125, 32, 125, 0, 0, 70, 98, 111, 108, 0, 1, 29, 5, 105, 110, 116, 38, 32, 88, 58, 58, 104, 97, 110, 100, 108, 101, 77, 111, 100, 117, 108, 97, 116, 105, 111, 110, 58, 58, 118, 0, 0, 82, 101, 116, 117, 114, 97, 116, 117, 114, 101,
			0, 1, 33, 5, 105, 110, 116, 32, 88, 58, 58, 104, 97, 110, 100, 108, 101, 77, 111, 100, 117, 108, 97, 116, 105, 111, 110, 40, 105, 110, 116, 38, 32, 118, 41, 0, 1, 1, 83, 121, 110, 97, 116, 117, 114, 101, 0, 1, 24, 5, 118, 111, 105, 100, 32, 115, 101, 116, 84, 111,
			70, 111, 117, 114, 40, 105, 110, 116, 38, 32, 114, 41, 0, 70, 117, 110, 99, 80, 97, 116, 117, 114, 101, 0, 1, 35, 5, 118, 111, 105, 100, 32, 115, 101, 116, 84, 111, 70, 111, 117, 114, 40, 105, 110, 116, 38, 32, 115, 101, 116, 84, 111, 70, 111, 117, 114, 58, 58, 114, 41,
			0, 67, 97, 108, 108, 84, 5, 123, 32, 123, 32, 49, 49, 48, 32, 125, 44, 32, 123, 32, 49, 46, 102, 32, 125, 44, 32, 123, 32, 52, 46, 102, 32, 125, 44, 32, 123, 32, 56, 46, 102, 32, 125, 44, 32, 123, 32, 57, 48, 32, 125, 32, 125, 0, 0, 70, 105, 110, 116, 44,
			32, 52, 48, 62, 0, 73, 100, 115, 0, 1, 10, 5, 88, 58, 58, 100, 97, 116, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 8, 5, 123, 32, 57, 48, 32, 125, 0, 0, 70, 117, 110, 116, 101, 114, 0, 1, 9, 6, 128, 160, 62, 12, 1, 0, 0,
			0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 65, 115, 115, 105, 103, 116, 84, 121, 112, 101, 0, 1, 4, 5, 88, 38, 0, 80, 97, 114, 101, 110, 116, 84, 121, 112, 101, 0, 1, 8,
			5, 77, 121, 83, 112, 97, 110, 0, 1, 2, 86, 97, 114, 105, 97, 116, 117, 114, 101, 0, 1, 31, 5, 118, 111, 105, 100, 32, 114, 101, 115, 101, 116, 79, 98, 106, 101, 99, 116, 40, 77, 121, 83, 112, 97, 110, 32, 99, 111, 112, 121, 41, 0, 70, 117, 110, 99, 80, 97, 116,
			117, 114, 101, 0, 1, 13, 5, 105, 110, 116, 32, 111, 116, 104, 101, 114, 40, 41, 0, 67, 97, 108, 108, 84, 121, 112, 101, 0, 1, 16, 5, 71, 108, 111, 98, 97, 108, 70, 117, 110, 99, 116, 105, 111, 110, 0, 0, 83, 98, 111, 108, 0, 1, 14, 5, 105, 110, 116, 38, 32,
			109, 97, 105, 110, 58, 58, 118, 0, 0, 65, 115, 115, 105, 103, 116, 44, 32, 50, 62, 38, 0, 80, 97, 114, 101, 110, 116, 84, 121, 112, 101, 0, 1, 23, 5, 115, 112, 97, 110, 60, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32, 50, 62, 44, 32, 50, 62, 0, 1, 2, 68,
			111, 116, 0, 97, 108, 117, 101, 115, 0, 1, 24, 5, 123, 32, 123, 32, 123, 32, 48, 51, 32, 125, 32, 125, 44, 32, 123, 32, 49, 51, 32, 125, 32, 125, 0, 0, 67, 88, 0, 73, 100, 115, 0, 1, 10, 5, 109, 97, 105, 110, 58, 58, 105, 44, 0, 73, 110, 105, 116, 86,
			97, 108, 117, 101, 115, 0, 1, 12, 5, 123, 32, 123, 32, 49, 51, 32, 125, 32, 125, 0, 0, 82, 116, 32, 115, 117, 109, 0, 0, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 67, 108, 97, 115,
			115, 83, 116, 5, 118, 111, 105, 100, 0, 0, 70, 117, 110, 99, 116, 121, 109, 98, 111, 108, 0, 1, 9, 5, 105, 110, 116, 32, 115, 117, 109, 0, 0, 65, 115, 115, 105, 103, 98, 111, 108, 0, 1, 28, 5, 115, 112, 97, 110, 60, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32,
			52, 62, 44, 32, 50, 62, 32, 100, 97, 116, 97, 0, 121, 109, 98, 111, 108, 0, 1, 35, 5, 105, 110, 116, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 50, 56, 58, 58, 105, 0, 0, 67, 111, 116, 84,
			121, 112, 101, 0, 1, 15, 5, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32, 52, 62, 38, 0, 1, 2, 83, 117, 98, 115, 101, 76, 105, 110, 101, 50, 52, 58, 58, 106, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0,
			0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 52, 0, 0, 83, 116, 0, 84, 121, 112, 101, 0, 1, 23, 5, 115, 112, 97, 110, 60, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32, 52, 62, 44, 32, 50, 62, 0, 73, 100, 98, 111, 108, 0, 1, 19, 5, 115, 112, 97,
			110, 60, 105, 110, 116, 44, 32, 55, 62, 32, 100, 97, 116, 97, 0, 0, 86, 97, 114, 105, 97, 0, 84, 121, 112, 101, 0, 1, 21, 5, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 60, 48, 44, 32, 48, 62, 0, 73, 100, 115, 0, 1, 6, 5, 105, 100, 120,
			44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 0, 1, 6, 5, 57, 48, 46, 102, 0, 0, 83, 117, 98, 115, 99, 114, 105, 0, 84, 121, 112, 101, 0, 1, 8, 5, 102, 108, 111, 97, 116, 52, 0, 73, 100, 115, 0, 1, 4, 5, 100, 44, 0, 73, 110, 105, 116, 86, 97,
			108, 117, 101, 115, 0, 1, 10, 5, 123, 32, 49, 57, 46, 102, 32, 125, 0, 0, 98, 111, 108, 0, 1, 36, 5, 102, 108, 111, 97, 116, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 115, 0,
			0, 86, 97, 114, 105, 97, 0, 1, 3, 5, 45, 0, 1, 2, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 4, 5, 49, 46, 0, 0, 86, 97, 114, 105, 97, 98, 111, 108, 0,
			1, 44, 5, 102, 108, 111, 97, 116, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 116, 104, 105, 115, 86, 97, 108, 117, 101, 0, 97, 114, 103, 101, 116, 0, 1, 7, 5, 102, 108, 111, 97, 116,
			0, 1, 1, 86, 97, 114, 105, 98, 111, 108, 0, 1, 23, 5, 102, 108, 111, 97, 116, 32, 109, 97, 105, 110, 58, 58, 108, 97, 115, 116, 86, 97, 108, 117, 101, 0, 0, 76, 111, 111, 112, 0, 117, 114, 101, 0, 1, 25, 5, 98, 108, 111, 99, 107, 32, 109, 97, 105, 110, 40,
			98, 108, 111, 99, 107, 32, 105, 110, 112, 117, 116, 41, 101, 114, 0, 1, 9, 6, 64, 160, 62, 12, 1, 0, 0, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 73, 102, 83, 116, 97,
			116, 101, 98, 111, 108, 0, 1, 12, 5, 105, 110, 116, 32, 118, 97, 108, 117, 101, 51, 0, 0, 82, 101, 116, 117, 114, 121, 112, 101, 0, 1, 7, 5, 98, 108, 111, 99, 107, 0, 73, 100, 115, 0, 1, 10, 5, 109, 97, 105, 110, 58, 58, 100, 44, 0, 73, 110, 105, 116, 86,
			97, 108, 117, 101, 115, 0, 1, 10, 5, 123, 32, 115, 44, 32, 56, 32, 102, 108, 111, 97, 116, 44, 32, 56, 62, 0, 73, 100, 115, 0, 1, 4, 5, 115, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 9, 5, 123, 32, 49, 46, 102, 32, 125, 0, 0, 70,
			117, 110, 98, 111, 108, 0, 1, 20, 5, 115, 112, 97, 110, 60, 88, 44, 32, 50, 62, 32, 109, 97, 105, 110, 58, 58, 120, 0, 0, 73, 109, 109, 101, 100, 105, 110, 58, 58, 120, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 21, 5, 123, 32, 123, 32, 105,
			50, 46, 102, 32, 125, 44, 32, 123, 32, 102, 32, 125, 32, 125, 0, 1, 3, 86, 97, 114, 105, 97, 98, 108, 98, 111, 108, 0, 1, 13, 5, 102, 108, 111, 97, 116, 32, 88, 58, 58, 102, 49, 0, 0, 65, 115, 115, 105, 121, 109, 98, 111, 108, 0, 1, 12, 5, 88, 58, 58,
			89, 32, 88, 58, 58, 121, 49, 0, 0, 67, 97, 115, 116, 0, 1, 3, 1, 8, 5, 88, 58, 58, 121, 49, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 21, 5, 123, 32, 123, 32, 49, 50, 46, 32, 125, 44, 32, 123, 32, 50, 46, 32, 125, 32, 125, 0,
			0, 65, 115, 115, 105, 103, 98, 111, 108, 0, 1, 17, 5, 100, 111, 117, 98, 108, 101, 32, 88, 58, 58, 89, 58, 58, 100, 49, 0, 0, 65, 115, 115, 105, 103, 101, 67, 97, 108, 108, 0, 1, 1, 67, 97, 115, 116, 0, 1, 3, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0,
			0, 0, 83, 111, 117, 114, 99, 101, 0, 1, 5, 5, 105, 110, 116, 0, 84, 97, 114, 103, 101, 116, 0, 1, 8, 121, 109, 98, 111, 108, 0, 1, 7, 5, 105, 110, 116, 32, 120, 0, 0, 65, 115, 115, 105, 103, 121, 109, 98, 111, 108, 0, 1, 40, 5, 72, 105, 115, 101, 69,
			118, 101, 110, 116, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 101, 0, 0, 86, 117, 114, 101, 0, 1, 29, 5, 105, 110, 116, 32, 72, 105, 115, 101, 69, 118, 101, 110, 116, 58, 58, 103, 101,
			116, 67, 104, 97, 110, 110, 101, 108, 40, 41, 0, 67, 97, 108, 108, 121, 109, 98, 111, 108, 0, 1, 28, 5, 80, 114, 111, 99, 101, 115, 115, 68, 97, 116, 97, 60, 49, 62, 38, 32, 109, 97, 105, 110, 58, 58, 100, 97, 116, 97, 0, 0, 0, 0, 0, 1, 1, 67, 111, 109,
			112, 97, 114, 105, 115, 111, 110, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 4, 5, 61, 61, 0, 1, 2, 86, 97, 114, 105, 117, 114, 101, 0, 1, 32, 5, 105, 110, 116, 32, 109, 97, 105, 110, 40, 80, 114, 111,
			99, 101, 115, 115, 68, 97, 116, 97, 60, 49, 62, 38, 32, 100, 97, 116, 97, 41, 0, 70, 117, 110, 99, 80, 97, 116, 117, 114, 101, 0, 1, 44, 5, 72, 105, 115, 101, 69, 118, 101, 110, 116, 66, 117, 102, 102, 101, 114, 32, 80, 114, 111, 99, 101, 115, 115, 68, 97, 116, 97,
			58, 58, 116, 111, 69, 118, 101, 110, 116, 68, 97, 116, 97, 40, 98, 111, 108, 0, 1, 20, 5, 100, 111, 117, 98, 108, 101, 32, 109, 97, 105, 110, 58, 58, 105, 110, 112, 117, 116, 0, 0, 82, 101, 116, 117, 114, 111, 108, 0, 1, 39, 5, 100, 111, 117, 98, 108, 101, 32, 112,
			114, 111, 99, 101, 115, 115, 111, 114, 58, 58, 115, 101, 116, 80, 97, 114, 97, 109, 101, 116, 101, 114, 58, 58, 118, 97, 108, 117, 101, 0, 0, 73, 109, 109, 101, 100, 117, 114, 101, 0, 1, 14, 5, 105, 110, 116, 32, 88, 58, 58, 103, 101, 116, 40, 41, 0, 67, 97, 108, 97,
			116, 117, 114, 101, 0, 1, 18, 5, 105, 110, 116, 32, 103, 101, 116, 65, 114, 103, 115, 60, 57, 62, 40, 41, 0, 70, 117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 0, 1, 9, 6, 0, 240, 69, 6, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 2, 84, 101, 109, 112, 108,
			97, 116, 101, 100, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 1, 76, 105, 110, 101, 0, 84, 121, 112, 101, 0, 1, 12, 5, 115, 112, 97, 110, 60, 88, 44, 32, 56, 62, 0, 73, 100, 115, 0, 1, 7, 5, 100, 97, 116, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117,
			101, 115, 0, 1, 10, 5, 123, 32, 123, 121, 109, 98, 111, 108, 0, 1, 33, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 105, 100, 120, 0, 0, 83, 121, 110, 116, 97, 108, 0, 1, 41,
			5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 43, 43, 111, 112, 101, 114, 97, 116, 111, 114, 58, 58, 110, 101, 119, 86, 97, 108, 117, 101, 0, 0, 65, 115, 115, 105, 121, 109, 98, 111, 108, 0, 1, 23, 5, 105, 110, 100, 101, 120,
			58, 58, 117, 110, 115, 97, 102, 101, 60, 48, 44, 32, 48, 62, 32, 105, 0, 0, 65, 115, 115, 105, 121, 109, 98, 111, 108, 0, 1, 26, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 118, 97, 108, 117, 101, 0, 0, 86, 97, 114,
			105, 97, 101, 73, 100, 0, 1, 24, 5, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 103, 101, 116, 70, 114, 111, 109, 0, 1, 5, 73, 110, 108, 105, 110, 101, 100, 65, 114, 103, 117, 109, 101, 110, 116, 0, 1, 3, 97, 108, 117, 101, 115, 0, 1, 34,
			5, 123, 32, 49, 46, 102, 44, 32, 50, 46, 102, 44, 32, 51, 46, 102, 44, 32, 52, 46, 102, 44, 32, 53, 46, 102, 44, 32, 54, 46, 102, 32, 125, 0, 0, 67, 121, 109, 98, 111, 108, 0, 1, 36, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112,
			112, 101, 100, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 108, 105, 109, 105, 116, 0, 0, 65, 115, 115, 105, 103, 121, 109, 98, 111, 108, 0, 1, 30, 5, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 60, 48, 44, 32, 48, 62, 32, 109, 97, 105, 110,
			58, 58, 105, 0, 0, 65, 115, 115, 105, 103, 98, 111, 108, 0, 1, 17, 5, 117, 110, 114, 101, 115, 111, 108, 118, 101, 100, 32, 116, 104, 105, 115, 0, 0, 68, 111, 116, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 73, 110, 108, 105, 110, 101,
			1, 5, 1, 0, 0, 0, 0, 65, 114, 103, 0, 1, 5, 1, 0, 0, 0, 0, 80, 97, 114, 97, 109, 101, 116, 101, 114, 78, 97, 109, 101, 0, 1, 40, 5, 105, 110, 116, 32, 105, 110, 0, 84, 121, 112, 101, 0, 1, 14, 5, 115, 112, 97, 110, 60, 105, 110, 116, 44, 32,
			53, 62, 0, 73, 100, 115, 0, 1, 4, 5, 99, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 19, 5, 123, 32, 49, 44, 62, 32, 100, 97, 116, 97, 50, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1,
			0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 48, 0, 0, 83, 117, 98, 115, 99, 100, 97, 116, 97, 49, 0, 0, 66, 105, 110, 97, 114, 121, 79, 112, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3,
			5, 43, 0, 1, 2, 70, 117, 110, 99, 121, 109, 98, 111, 108, 0, 1, 16, 5, 100, 121, 110, 60, 105, 110, 116, 62, 32, 100, 97, 116, 97, 49, 0, 0, 86, 97, 114, 105, 97, 97, 108, 117, 101, 115, 0, 1, 35, 5, 123, 32, 49, 44, 32, 50, 44, 32, 51, 44, 32, 52,
			44, 32, 53, 44, 32, 54, 44, 32, 55, 44, 32, 56, 44, 32, 57, 44, 32, 49, 48, 32, 125, 0, 0, 67, 111, 111, 112, 101, 73, 100, 0, 1, 28, 5, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 50, 52,
			0, 1, 1, 65, 115, 115, 105, 103, 110, 109, 0, 84, 121, 112, 101, 0, 1, 17, 5, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 49, 53, 62, 0, 73, 100, 115, 0, 1, 4, 5, 115, 44, 0, 73, 110, 105, 116, 86, 116, 84, 121, 112, 101, 0, 1, 6, 5, 105,
			110, 116, 38, 0, 80, 97, 114, 101, 110, 116, 84, 121, 112, 101, 0, 1, 10, 5, 100, 121, 110, 60, 105, 110, 116, 62, 0, 1, 2, 86, 97, 114, 105, 97, 116, 117, 114, 101, 0, 1, 108, 5, 100, 121, 110, 60, 105, 110, 116, 62, 32, 100, 121, 110, 58, 58, 114, 101, 102, 101,
			114, 84, 111, 40, 117, 101, 0, 1, 4, 5, 49, 53, 0, 0, 68, 111, 116, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 68, 111, 116, 0, 1, 1, 76, 97, 108, 117, 101, 115, 0, 1, 16, 5, 123, 32, 123, 32, 123, 32, 49, 50, 32, 125,
			32, 125, 32, 125, 0, 0, 70, 117, 110, 121, 109, 98, 111, 108, 0, 1, 11, 5, 88, 58, 58, 89, 32, 88, 58, 58, 121, 0, 0, 86, 97, 114, 105, 97, 121, 112, 101, 0, 1, 6, 5, 88, 58, 58, 89, 0, 73, 100, 115, 0, 1, 7, 5, 88, 58, 58, 121, 44, 0, 73,
			110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 12, 5, 123, 32, 123, 32, 49, 57, 32, 125, 32, 125, 0, 121, 109, 98, 111, 108, 0, 1, 19, 5, 98, 108, 111, 99, 107, 32, 109, 97, 105, 110, 58, 58, 105, 110, 112, 117, 116, 0, 0, 83, 121, 110, 116, 97, 105, 110, 112,
			117, 116, 44, 32, 102, 108, 111, 97, 116, 32, 108, 111, 119, 101, 114, 108, 105, 109, 105, 116, 44, 32, 102, 108, 111, 97, 116, 32, 117, 112, 112, 101, 114, 108, 105, 109, 105, 116, 41, 0, 67, 97, 108, 108, 98, 111, 108, 0, 1, 34, 5, 88, 32, 109, 97, 105, 110, 58, 58, 65,
			110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 111, 98, 106, 0, 0, 82, 101, 116, 117, 114, 111, 112, 101, 76, 105, 110, 101, 51, 53, 58, 58, 111, 98, 106, 50, 44, 0, 73, 110, 105, 116, 86, 111, 112, 101, 73, 100, 0, 1, 49,
			5, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 65, 110, 111, 110, 121, 97, 116, 117, 114, 101, 0, 1, 29, 5, 100, 111, 117, 98, 108, 101, 32, 77, 97, 116, 104, 58, 58, 114, 97, 110, 100, 111,
			109, 68, 111, 117, 98, 108, 101, 40, 41, 0, 67, 97, 108, 1, 3, 5, 42, 0, 1, 2, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 50, 0, 0, 86, 97, 114, 105,
			97, 97, 116, 117, 114, 101, 0, 1, 14, 5, 118, 111, 105, 100, 32, 88, 58, 58, 126, 88, 40, 41, 0, 67, 97, 108, 88, 0, 73, 100, 115, 0, 1, 33, 5, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48,
			58, 58, 111, 98, 106, 44, 0, 121, 112, 101, 0, 1, 3, 5, 60, 0, 1, 2, 86, 97, 114, 105, 0, 0, 0, 1, 2, 73, 110, 99, 114, 101, 109, 101, 110, 116, 0, 1, 3, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 73, 115, 80, 114, 101, 0, 1, 1, 2,
			73, 115, 68, 101, 99, 0, 1, 108, 117, 101, 0, 1, 3, 5, 53, 0, 0, 99, 111, 110, 116, 105, 110, 117, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 1, 70, 117, 110, 99, 117, 114, 101, 0, 1, 13, 5, 118, 111, 105, 100, 32, 88, 58,
			58, 88, 40, 41, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 73, 110, 121, 109, 98, 111, 108, 0, 1, 29, 5, 115, 112, 97, 110, 60, 68, 101, 108, 97, 121, 66, 117, 102, 102, 101,
			114, 44, 32, 50, 62, 32, 98, 117, 102, 102, 101, 114, 0, 0, 73, 109, 109, 101, 100, 101, 114, 44, 32, 50, 62, 0, 73, 100, 115, 0, 1, 9, 5, 98, 117, 102, 102, 101, 114, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 13, 5, 123, 32, 123, 32, 48,
			46, 102, 32, 125, 32, 125, 0, 0, 67, 97, 116, 117, 114, 101, 0, 1, 17, 5, 105, 110, 116, 32, 100, 121, 110, 58, 58, 115, 105, 122, 101, 40, 41, 0, 67, 97, 108, 108, 102, 101, 114, 84, 111, 58, 58, 115, 105, 122, 101, 44, 32, 105, 110, 116, 32, 100, 121, 110, 58, 58,
			114, 101, 102, 101, 114, 84, 111, 58, 58, 111, 102, 102, 115, 101, 116, 41, 0, 67, 97, 108, 108, 0, 84, 121, 112, 101, 0, 1, 10, 5, 100, 121, 110, 60, 105, 110, 116, 62, 0, 73, 100, 115, 0, 1, 8, 5, 100, 97, 116, 97, 49, 44, 0, 73, 110, 105, 116, 86, 0, 84,
			121, 112, 101, 0, 1, 7, 5, 102, 108, 111, 97, 116, 0, 1, 1, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 5, 5, 97, 116, 117, 114, 101, 0, 1, 23, 5, 105, 110, 116,
			32, 103, 101, 116, 68, 101, 102, 97, 117, 108, 116, 40, 105, 110, 116, 32, 120, 41, 0, 70, 117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 0, 1, 9, 6, 0, 16, 66, 6, 121, 109, 98, 111, 108, 0, 1, 38, 5, 83, 99, 111, 112, 101, 100, 88, 32, 109, 97, 105, 110, 58,
			58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 97, 0, 0, 65, 115, 115, 108, 117, 101, 0, 1, 3, 5, 48, 0, 0, 86, 97, 114, 105, 5, 105, 110, 116, 0, 1, 1, 66, 105, 110, 97, 114, 121, 79, 112, 0, 1, 2,
			76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3, 5, 43, 0, 1, 2, 86, 97, 114, 105, 101, 48, 58, 58, 105, 0, 0, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0,
			0, 86, 97, 108, 117, 101, 0, 1, 4, 5, 49, 52, 0, 0, 83, 116, 97, 116, 101, 101, 114, 0, 1, 9, 6, 0, 17, 66, 6, 1, 0, 0, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0,
			1, 3, 65, 115, 115, 105, 103, 112, 101, 100, 88, 58, 58, 83, 99, 111, 112, 101, 100, 88, 40, 41, 0, 67, 97, 108, 108, 84, 121, 112, 101, 0, 1, 16, 5, 77, 101, 109, 98, 101, 114, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 1, 77, 101, 109, 111, 114, 121, 109, 98,
			111, 108, 0, 1, 20, 5, 105, 110, 116, 32, 83, 99, 111, 112, 101, 100, 88, 58, 58, 118, 97, 108, 117, 101, 0, 0, 67, 108, 97, 115, 115, 83, 116, 1, 0, 0, 0, 0, 1, 3, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0,
			0, 0, 0, 83, 105, 103, 110, 97, 116, 117, 114, 101, 0, 1, 25, 5, 118, 111, 105, 100, 32, 83, 121, 109, 98, 111, 108, 0, 1, 15, 5, 105, 110, 116, 32, 109, 97, 105, 110, 58, 58, 115, 117, 109, 0, 0, 70, 117, 110, 99, 116, 121, 109, 98, 111, 108, 0, 1, 34, 5,
			105, 110, 116, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 105, 0, 0, 67, 111, 109, 112, 97, 114, 105, 121, 109, 98, 111, 108, 0, 1, 14, 5, 79, 117, 116, 101, 114, 32, 116, 117, 116, 58,
			58, 101, 0, 0, 70, 117, 110, 99, 116, 105, 111, 110, 67, 97, 108, 108, 0, 1, 3, 76, 105, 110, 101, 99, 111, 112, 101, 100, 88, 0, 73, 100, 115, 0, 1, 11, 5, 79, 117, 116, 101, 114, 58, 58, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 11,
			5, 123, 32, 123, 32, 57, 32, 125, 32, 125, 0, 0, 67, 114, 101, 0, 1, 12, 5, 118, 111, 105, 100, 32, 116, 117, 116, 40, 41, 0, 67, 97, 108, 108, 84, 121, 112, 101, 0, 1, 16, 5, 71, 108, 111, 98, 97, 108, 70, 117, 110, 99, 116, 105, 111, 110, 0, 0, 82, 101,
			116, 117, 114, 121, 112, 101, 0, 1, 7, 5, 79, 117, 116, 101, 114, 0, 77, 101, 109, 98, 101, 114, 84, 121, 112, 101, 0, 1, 9, 5, 83, 99, 111, 112, 101, 100, 88, 0, 79, 102, 102, 115, 101, 116, 0, 1, 5, 1, 4, 0, 0, 0, 1, 116, 101, 114, 0, 73, 100, 115,
			0, 1, 9, 5, 116, 117, 116, 58, 58, 101, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 26, 5, 123, 32, 123, 32, 123, 32, 57, 32, 125, 32, 125, 44, 32, 123, 32, 123, 32, 97, 116, 117, 114, 101, 0, 1, 26, 5, 118, 111, 105, 100, 32, 83, 99, 111,
			112, 101, 100, 88, 58, 58, 126, 83, 99, 111, 112, 101, 100, 88, 40, 41, 0, 1, 1, 83, 121, 1, 13, 5, 105, 110, 116, 32, 99, 111, 117, 110, 116, 101, 114, 0, 0, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0,
			0, 0, 1, 5, 65, 115, 115, 105, 103, 110, 109, 117, 114, 101, 0, 1, 15, 5, 118, 111, 105, 100, 32, 99, 104, 97, 110, 103, 101, 40, 41, 0, 70, 117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 0, 1, 9, 6, 0, 160, 62, 12, 1, 0, 0, 0, 1, 1, 83, 0, 0,
			73, 110, 99, 114, 101, 109, 101, 110, 116, 0, 1, 3, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 73, 115, 80, 114, 101, 0, 1, 1, 3, 73, 115, 68, 101, 99, 0, 1, 1, 3, 1, 1, 86, 97, 114, 105, 1, 0, 0, 0, 0, 1, 2, 87, 104, 105, 108, 101,
			76, 111, 111, 112, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 4, 65, 115, 115, 105, 103, 110, 109, 1, 13, 5, 105, 110, 116, 32, 109, 97, 105, 110, 58, 58, 118, 0, 0, 65, 110, 111, 110, 121, 109, 111, 117, 115, 66, 108, 111, 99, 107, 0, 1,
			1, 76, 105, 110, 101, 98, 111, 108, 0, 1, 13, 5, 105, 110, 116, 32, 88, 58, 58, 88, 58, 58, 97, 0, 0, 68, 111, 116, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 84, 104, 105, 115, 80, 111, 105, 112, 101, 114, 116, 121, 0, 1, 3,
			76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 73, 68, 0, 1, 13, 5, 78, 117, 109, 67, 104, 97, 110, 110, 101, 108, 115, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 51, 0, 0, 121, 109, 98, 111, 108, 0, 1, 31, 5, 115, 112, 97, 110, 60, 105, 110, 116, 44,
			32, 49, 51, 62, 38, 32, 103, 101, 116, 70, 105, 114, 115, 116, 58, 58, 100, 97, 116, 97, 0, 0, 73, 109, 109, 101, 100, 105, 97, 58, 73, 110, 110, 101, 114, 58, 58, 103, 101, 116, 40, 41, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105,
			110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 1, 82, 101, 116, 117, 114, 110, 83, 117, 116, 101, 114, 58, 58, 73, 110, 110, 101, 114, 60, 51, 62, 0, 73, 100, 115, 0, 1, 4, 5, 100, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 6, 5, 123, 32,
			32, 125, 0, 0, 70, 117, 110, 99, 116, 121, 109, 98, 111, 108, 0, 1, 13, 5, 88, 32, 109, 97, 105, 110, 58, 58, 111, 98, 106, 0, 0, 83, 121, 110, 116, 97, 120, 84, 1, 12, 5, 109, 97, 105, 110, 58, 58, 111, 98, 106, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117,
			101, 115, 0, 1, 11, 5, 123, 32, 123, 32, 53, 32, 125, 32, 125, 0, 0, 82, 101, 116, 117, 114, 5, 105, 110, 116, 0, 1, 1, 84, 101, 114, 110, 97, 114, 121, 79, 112, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 3, 67, 111, 109, 112, 97,
			114, 105, 114, 97, 116, 111, 114, 0, 1, 37, 5, 102, 108, 111, 97, 116, 38, 32, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 58, 58, 115, 0, 1, 2, 86, 97, 114, 48, 32, 125, 32, 125, 0, 0, 73,
			102, 83, 116, 97, 116, 101, 109, 101, 110, 116, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 67, 111, 109, 112, 97, 114, 105, 115, 111, 110, 0, 1, 2, 76, 105, 1, 3, 5, 61, 0, 1, 2, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1,
			2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 4, 5, 57, 48, 0, 0, 86, 97, 114, 105, 97, 116, 121, 44, 32, 100, 99, 44, 32, 100, 99, 62, 0, 77, 101, 109, 98, 101, 114, 84, 121, 112, 101, 0, 1, 5, 5, 100, 99, 38,
			0, 79, 102, 102, 115, 101, 116, 0, 1, 5, 1, 0, 0, 0, 0, 1, 1, 86, 97, 114, 105, 116, 105, 111, 110, 0, 1, 2, 77, 101, 109, 111, 114, 121, 82, 101, 102, 101, 114, 101, 110, 99, 101, 0, 1, 4, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 98,
			106, 101, 99, 116, 84, 121, 112, 101, 0, 1, 44, 1, 5, 1, 0, 0, 0, 0, 73, 68, 0, 1, 8, 5, 73, 115, 78, 111, 100, 101, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 49, 0, 0, 73, 110, 116, 101, 114, 110, 97, 108, 80, 114, 111, 112, 101, 114, 116, 121, 0,
			1, 121, 109, 98, 111, 108, 0, 1, 24, 5, 102, 108, 111, 97, 116, 32, 77, 117, 108, 116, 105, 112, 108, 105, 101, 114, 58, 58, 103, 97, 105, 110, 0, 0, 76, 111, 111, 112, 0, 1, 4, 76, 105, 110, 101, 3, 69, 108, 101, 109, 101, 110, 116, 84, 121, 112, 101, 0, 1, 13,
			5, 77, 117, 108, 116, 105, 112, 108, 105, 101, 114, 38, 0, 80, 97, 114, 101, 110, 116, 84, 121, 112, 101, 0, 1, 21, 5, 115, 112, 97, 110, 60, 77, 117, 108, 97, 116, 117, 114, 101, 0, 1, 25, 5, 102, 108, 111, 97, 116, 32, 109, 97, 105, 110, 40, 102, 108, 111, 97, 116,
			32, 105, 110, 112, 117, 116, 41, 0, 70, 117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 0, 1, 9, 6, 64, 240, 0, 84, 121, 112, 101, 0, 1, 5, 5, 105, 110, 116, 0, 1, 1, 68, 111, 116, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2,
			86, 97, 114, 105, 97, 98, 108, 101, 0, 1, 3, 5, 88, 0, 73, 100, 115, 0, 1, 4, 5, 120, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 30, 5, 123, 32, 123, 32, 49, 50, 46, 102, 32, 125, 44, 32, 123, 32, 57, 32, 125, 44, 32, 97, 116, 117,
			114, 101, 0, 1, 32, 5, 100, 111, 117, 98, 108, 101, 32, 77, 97, 116, 104, 58, 58, 97, 98, 115, 40, 100, 111, 117, 98, 108, 101, 32, 105, 110, 112, 117, 116, 41, 0, 67, 97, 108, 108, 84, 111, 117, 114, 99, 101, 0, 1, 8, 5, 100, 111, 117, 98, 108, 101, 0, 84, 97,
			114, 103, 101, 116, 0, 1, 5, 5, 105, 110, 116, 0, 1, 1, 70, 117, 110, 99, 116, 105, 111, 110, 67, 97, 108, 108, 0, 1, 3, 76, 105, 110, 97, 116, 117, 114, 101, 0, 1, 15, 5, 102, 108, 111, 97, 116, 38, 32, 103, 101, 116, 88, 40, 41, 0, 67, 97, 108, 108, 84,
			121, 112, 101, 0, 1, 16, 5, 71, 108, 111, 98, 97, 108, 70, 117, 110, 99, 116, 105, 111, 121, 109, 98, 111, 108, 0, 1, 10, 5, 105, 110, 116, 32, 88, 58, 58, 118, 0, 0, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 3, 76, 105, 110, 101, 0, 84, 121, 112, 101, 0,
			1, 3, 5, 88, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 2, 70, 117, 110, 99, 116, 105, 111, 109, 97, 105, 110, 40, 102, 108, 111, 97, 116, 32, 105, 110, 112, 117, 116, 41, 0, 70,
			117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 0, 1, 9, 6, 0, 240, 153, 9, 1, 0, 0, 0, 1, 1, 83, 121, 110, 116, 97, 120, 84, 101, 102, 101, 114, 84, 111, 40, 99, 111, 110, 115, 116, 32, 112, 111, 105, 110, 116, 101, 114, 32, 100, 121, 110, 58, 58, 114, 101, 102,
			101, 114, 84, 111, 58, 58, 111, 116, 104, 101, 114, 44, 32, 105, 110, 116, 32, 100, 121, 110, 58, 118, 101, 83, 121, 110, 116, 104, 58, 58, 112, 104, 97, 115, 101, 66, 108, 111, 99, 107, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 12, 5, 123, 32, 48, 120,
			48, 44, 32, 48, 32, 125, 0, 0, 67, 111, 109, 115, 111, 108, 118, 101, 100, 32, 105, 110, 100, 101, 120, 58, 58, 117, 110, 115, 97, 102, 101, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 99, 0, 1, 1, 68, 111, 116, 0, 1, 1, 76, 105, 110, 101, 112, 84, 121, 112,
			101, 0, 1, 7, 5, 66, 108, 111, 99, 107, 0, 76, 111, 97, 100, 73, 116, 101, 114, 97, 116, 111, 114, 0, 1, 1, 2, 73, 116, 101, 114, 97, 116, 111, 114, 0, 1, 52, 5, 102, 108, 111, 97, 116, 38, 32, 97, 116, 117, 114, 101, 0, 1, 43, 5, 105, 110, 116, 32, 77,
			97, 116, 104, 58, 58, 109, 97, 120, 40, 105, 110, 116, 32, 105, 110, 112, 117, 116, 44, 32, 105, 110, 116, 32, 109, 97, 120, 58, 58, 80, 97, 114, 97, 109, 49, 41, 97, 112, 112, 101, 100, 58, 58, 103, 101, 116, 70, 114, 111, 109, 58, 58, 105, 100, 120, 0, 0, 73, 110, 108,
			105, 110, 101, 100, 82, 101, 116, 117, 114, 110, 86, 97, 108, 117, 101, 0, 1, 2, 76, 105, 110, 97, 116, 117, 114, 101, 0, 1, 30, 5, 105, 110, 116, 32, 77, 97, 116, 104, 58, 58, 119, 114, 97, 112, 40, 105, 110, 116, 32, 97, 44, 32, 105, 110, 116, 32, 98, 41, 0, 67,
			97, 108, 108, 84, 60, 105, 110, 116, 44, 32, 50, 62, 38, 0, 73, 100, 115, 0, 1, 10, 5, 109, 97, 105, 110, 58, 58, 118, 44, 0, 1, 1, 73, 110, 108, 105, 110, 101, 100, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 2, 76, 105, 110, 101, 100, 60, 50, 44, 32, 48,
			62, 0, 0, 65, 115, 115, 105, 103, 110, 109, 101, 110, 116, 0, 1, 3, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 70, 105, 114, 115, 116, 0, 1, 1, 3, 65, 115, 115, 105, 103, 110, 109, 62, 38, 0, 1, 1, 80, 111, 105, 110, 116, 101, 114, 65, 99, 99, 101,
			115, 115, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 1, 84, 104, 105, 115, 80, 111, 105, 110, 116, 101, 114, 0, 1, 2, 76, 101, 73, 100, 0, 1, 27, 5, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58, 111, 112, 101,
			114, 97, 116, 111, 114, 61, 0, 1, 4, 73, 110, 108, 105, 110, 101, 100, 65, 114, 103, 117, 109, 101, 110, 116, 97, 112, 112, 101, 100, 60, 51, 44, 32, 48, 62, 0, 73, 100, 115, 0, 1, 4, 5, 105, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 11, 5,
			123, 32, 123, 32, 48, 32, 125, 32, 125, 0, 0, 67, 111, 109, 116, 84, 121, 112, 101, 0, 1, 8, 5, 102, 108, 111, 97, 116, 38, 0, 80, 97, 114, 101, 110, 116, 84, 121, 112, 101, 0, 1, 16, 5, 115, 112, 97, 110, 60, 102, 108, 111, 97, 116, 44, 32, 50, 62, 0, 1,
			2, 86, 97, 114, 121, 112, 101, 0, 1, 3, 5, 43, 0, 1, 2, 83, 117, 98, 115, 99, 114, 105, 112, 116, 0, 1, 4, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 87, 114, 105, 116, 101, 0, 1, 1, 3, 69, 108, 101, 109, 101, 110, 108, 117, 101, 0, 0, 82,
			101, 116, 117, 114, 110, 83, 116, 97, 116, 101, 109, 101, 110, 116, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 84, 121, 112, 101, 0, 1, 6, 5, 118, 111, 105, 100, 0, 0, 65, 0, 1, 2, 66, 105, 110, 97, 114, 121, 79, 112, 0, 1, 2, 76, 105,
			110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 79, 112, 84, 121, 112, 101, 0, 1, 3, 5, 42, 0, 1, 2, 67, 97, 115, 116, 0, 1, 3, 76, 105, 110, 112, 108, 101, 78, 111, 100, 101, 58, 58, 112, 114, 111, 99, 101, 115, 115, 70, 114, 97, 109, 101, 58, 58, 100, 97, 116,
			97, 0, 0, 83, 116, 97, 116, 101, 109, 101, 110, 116, 66, 108, 111, 99, 107, 0, 1, 2, 76, 105, 110, 101, 1, 5, 1, 0, 0, 0, 0, 76, 111, 111, 112, 84, 121, 112, 101, 0, 1, 6, 5, 83, 112, 97, 110, 0, 76, 111, 97, 100, 73, 116, 101, 114, 97, 116, 111, 114,
			0, 1, 1, 3, 73, 116, 101, 114, 97, 116, 111, 5, 1, 0, 0, 0, 0, 83, 99, 111, 112, 101, 73, 100, 0, 1, 27, 5, 109, 97, 105, 110, 58, 58, 65, 110, 111, 110, 121, 109, 111, 117, 115, 83, 99, 111, 112, 101, 76, 105, 110, 101, 48, 0, 1, 1, 65, 115, 115, 105,
			103, 44, 32, 49, 50, 56, 62, 0, 73, 100, 115, 0, 1, 7, 5, 100, 97, 116, 97, 44, 0, 73, 110, 105, 116, 86, 97, 108, 117, 101, 115, 0, 1, 9, 5, 123, 32, 48, 46, 102, 32, 125, 0, 0, 67, 111, 109, 112, 108, 101, 120, 114, 97, 109, 49, 41, 0, 67, 97, 108,
			108, 84, 121, 112, 101, 0, 1, 16, 5, 78, 97, 116, 105, 118, 101, 84, 121, 112, 101, 67, 97, 108, 108, 0, 1, 2, 73, 109, 109, 101, 100, 105, 97, 116, 101, 0, 1, 2, 76, 105, 110, 1, 5, 1, 0, 0, 0, 0, 83, 111, 117, 114, 99, 101, 0, 1, 7, 5, 102, 108,
			111, 97, 116, 0, 84, 97, 114, 103, 101, 116, 0, 1, 5, 5, 105, 110, 116, 0, 1, 1, 73, 110, 108, 105, 110, 101, 100, 70, 117, 110, 99, 97, 112, 112, 101, 100, 58, 58, 97, 115, 115, 105, 103, 110, 73, 110, 116, 101, 114, 110, 97, 108, 58, 58, 118, 0, 1, 1, 73, 110,
			108, 105, 110, 101, 100, 80, 97, 114, 97, 109, 101, 116, 101, 114, 0, 1, 2, 76, 105, 110, 101, 109, 98, 111, 108, 0, 1, 27, 5, 105, 110, 116, 32, 105, 110, 100, 101, 120, 58, 58, 119, 114, 97, 112, 112, 101, 100, 58, 58, 118, 97, 108, 117, 101, 0, 0, 73, 110, 108, 105,
			110, 101, 100, 82, 101, 116, 117, 114, 110, 86, 97, 1, 0, 0, 0, 0, 65, 114, 103, 0, 1, 5, 1, 255, 255, 255, 255, 80, 97, 114, 97, 109, 101, 116, 101, 114, 78, 97, 109, 101, 0, 1, 17, 5, 117, 110, 114, 101, 115, 111, 108, 118, 101, 100, 32, 116, 104, 105, 115, 0,
			1, 58, 58, 103, 101, 116, 77, 49, 40, 41, 0, 67, 97, 108, 108, 84, 121, 112, 101, 0, 1, 16, 5, 77, 101, 109, 98, 101, 114, 70, 117, 110, 99, 116, 105, 111, 110, 0, 1, 1, 86, 97, 114, 105, 97, 98, 108, 5, 1, 0, 0, 0, 0, 83, 121, 109, 98, 111, 108, 0,
			1, 17, 5, 105, 110, 116, 32, 109, 97, 105, 110, 58, 58, 105, 110, 112, 117, 116, 0, 0, 70, 117, 110, 99, 116, 105, 111, 110, 67, 97, 108, 108, 0, 1, 3, 76, 105, 5, 1, 0, 0, 0, 0, 1, 2, 67, 111, 109, 112, 108, 101, 120, 84, 121, 112, 101, 68, 101, 102, 105,
			110, 105, 116, 105, 111, 110, 0, 1, 4, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 84, 121, 112, 101, 0, 1, 2, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 86, 97, 108, 117, 101, 0, 1, 3, 5, 57, 0, 0, 86, 97, 114, 105, 97, 98, 108, 101,
			82, 101, 102, 101, 114, 101, 110, 99, 101, 0, 1, 2, 76, 105, 110, 101, 83, 121, 110, 116, 97, 120, 84, 114, 101, 101, 0, 1, 1, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 1, 3, 67, 108, 97, 115, 115, 83, 116, 97, 116, 101, 109, 101, 110, 116, 0, 1, 2,
			76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 83, 105, 103, 110, 97, 116, 117, 114, 101, 0, 1, 21, 5, 105, 110, 116, 32, 109, 97, 105, 110, 40, 105, 110, 116, 32, 105, 110, 112, 117, 116, 41, 0, 70, 117, 110, 99, 80, 111, 105, 110, 116, 101, 114, 109, 101, 110, 116,
			0, 1, 3, 76, 105, 110, 101, 0, 1, 5, 1, 0, 0, 0, 0, 70, 105, 114, 115, 116, 0, 1, 1, 2, 65, 115, 115, 105, 103, 110, 109, 101, 110, 116, 84, 121, 112, 101, 0, 1, 3, 5, 61, 0, 1, 2 };

		return MemoryBlock(dictionary, sizeof(dictionary));
	}
};

juce::ValueTree SyntaxTreeExtractor::getSyntaxTree(const String& b64)
{
	ValueTree v;

	if (isBase64Tree(b64))
	{
		static const String prefix("B64Tree::");
		MemoryBlock mb;
		mb.fromBase64Encoding(b64.substring(prefix.length()));

		zstd::ZCompressor<SyntaxTreeDictionaryProvider> comp;

		auto ok = comp.expand(mb, v);
		jassert(ok.wasOk());
	}

	return v;
}

String SyntaxTreeExtractor::getBase64SyntaxTree(ValueTree v)
{
	removeLineInfo(v);

	String b64;

	if (v.isValid())
	{
		b64 << "B64Tree::";
		MemoryBlock mb;
		zstd::ZCompressor<SyntaxTreeDictionaryProvider> comp;
		auto ok = comp.compress(v, mb);
		jassert(ok);

		b64 << mb.toBase64Encoding();
	}

	return b64;
}

bool SyntaxTreeExtractor::isBase64Tree(const String& code)
{
	static const String prefix("B64Tree::");
	return code.startsWith(prefix);
}

void SyntaxTreeExtractor::removeLineInfo(ValueTree& v)
{
	v.removeProperty("Line", nullptr);
	v.removeProperty("FuncPointer", nullptr);

#if 0
	if (v.hasProperty("InitValues"))
	{
		InitValueParser p(v.getProperty("InitValues"));
		auto x = p.getB64();
		v.setProperty("InitValues", x, nullptr);
	}
#endif

	for (auto c : v)
		removeLineInfo(c);
}

String SyntaxTreeExtractor::getBase64DataLayout(const Array<ValueTree>& dataLayouts)
{
    MemoryOutputStream mos;
    
    for(const auto& x: dataLayouts)
    {
        zstd::ZDefaultCompressor comp;
        
        MemoryBlock mb;
        comp.compress(x, mb);
        
        mos.writeInt((int)mb.getSize());
        mos.write(mb.getData(), mb.getSize());
    }
    
    mos.flush();
    return "b64" + mos.getMemoryBlock().toBase64Encoding();
}

Array<juce::ValueTree> SyntaxTreeExtractor::getDataLayoutTrees(const juce::String &b64)
{
    Array<ValueTree> list;
    
    jassert(b64.startsWith("b64"));
    
    MemoryBlock mb;
    mb.fromBase64Encoding(b64.substring(3));
    MemoryInputStream mos(std::move(mb));
    
    while(!mos.isExhausted())
    {
        int numBytes = mos.readInt();
        
        MemoryBlock mb;
        mb.ensureSize(numBytes);
        mos.read(mb.getData(), numBytes);
        
        zstd::ZDefaultCompressor comp;
        
        ValueTree v;
        comp.expand(mb, v);
        
        list.add(v);
    }
    
    return list;
}


void InitValueParser::forEach(const std::function<void(uint32 offset, Types::ID type, const VariableStorage& value)>& f) const
{
	if (!input.startsWith("b64"))
	{
		InitValueParser p(getB64());
		p.forEach(f);
		return;
	}

	auto d = input.substring(3);

	MemoryBlock mb;
	mb.fromBase64Encoding(d);
	MemoryInputStream mos(std::move(mb));

	int offset = 0;

	while (!mos.isExhausted())
	{
		auto b = mos.readByte();

		switch (b)
		{
		case 'i':
			f(offset, Types::ID::Integer, VariableStorage(mos.readInt()));
			offset += sizeof(int);
			break;
		case 'f':
			f(offset, Types::ID::Float, VariableStorage(mos.readFloat()));
			offset += sizeof(float);
			break;
		case 'd': 
			
			if (offset % sizeof(double) != 0)
				offset += sizeof(double) - (offset % sizeof(double));

			f(offset, Types::ID::Double, VariableStorage(mos.readDouble()));

			offset += sizeof(double);
			break;
		case 'p':

			if (offset % sizeof(void*) != 0)
				offset += sizeof(void*) - (offset % sizeof(void*));

			f(offset, Types::ID::Pointer, VariableStorage(reinterpret_cast<void*>(mos.readInt64()), 8));
			offset += sizeof(void*);
			break;
		case 'I':
			f(offset, Types::ID::Integer, VariableStorage((void*)(uint64_t)mos.readInt()));
			offset += sizeof(int);
			break;
		case 'F':
			f(offset, Types::ID::Float, VariableStorage((void*)(uint64_t)mos.readInt()));
			offset += sizeof(float);
			break;
		case 'D':

			if (offset % sizeof(double) != 0)
				offset += sizeof(double) - (offset % sizeof(double));

			f(offset, Types::ID::Double, VariableStorage((void*)(uint64_t)mos.readInt()));
			offset += sizeof(double);

			break;
		default: jassertfalse; return;
		}
	}
}

juce::uint32 InitValueParser::getNumBytesRequired() const
{
	uint32 b = 0;

	forEach([&b](uint32 offset, Types::ID t, const VariableStorage&)
	{
		auto nb = (uint32)Types::Helpers::getSizeForType(t);

		if (t == Types::ID::Integer)
			nb = sizeof(int64);

		b = offset + nb;
	});

	return b;
}

String InitValueParser::dumpContent() const
{
	String s;

	forEach([&s](int offset, Types::ID type, const VariableStorage& v)
	{
		auto typestring = Types::Helpers::getTypeName(type);

		if (v.getType() == Types::ID::Pointer)
			typestring = "expr";

		s << typestring << ":" << Types::Helpers::getCppValueString(v) << " ";
	});

	return s;
	//s << "float: " << mos.readFloat() << ", "; break;
}

String InitValueParser::getB64() const
{
	MemoryOutputStream mos;

	jassert(!input.startsWith("b64"));
	

	auto ptr = input.begin();
	auto end = input.end();

	while (ptr != end)
	{
		auto c = *ptr;

		if (CharacterFunctions::isWhitespace(c) ||
			c == '{' || c == '}' || c == ',')
		{
			ptr++;
			continue;
		}

		auto isExpression = false;

		uint8_t expressionType = 0;

		if (*ptr == '$')
		{
			isExpression = true;
			ptr++;
			expressionType = *ptr++;
		}

		bool isMinus = *ptr == '-';

		if (isMinus)
			ptr++;

		auto b = ptr;
		auto e = ptr;

		while (e != end &&
			    (CharacterFunctions::isDigit(*e) ||
				*e == '.' ||
				*e == 'f' ||
                *e == 'x' ||
			    *e == 'p'))
		{
			e++;
		}

		String value(b, e);

        if(value.isEmpty())
            break;
        
		if (isExpression)
		{
			mos.writeByte(CharacterFunctions::toUpperCase(expressionType));
			mos.writeInt(value.getIntValue());
		}
		else
		{
			auto type = Types::Helpers::getTypeFromStringValue(value);

			if (type == Types::ID::Integer)
			{
				mos.writeByte('i');
				mos.writeInt(value.getIntValue() * (!isMinus ? 1 : -1));
			}
			if (type == Types::ID::Float)
			{
				mos.writeByte('f');
				mos.writeFloat(value.getFloatValue() * (!isMinus ? 1.0f : -1.0f));
			}
			if (type == Types::ID::Double)
			{
				mos.writeByte('d');
				mos.writeDouble(value.getDoubleValue() * (!isMinus ? 1.0 : -1.0));
			}
			if (type == Types::ID::Pointer)
			{
				mos.writeByte('p');

				auto v2 = value.substring(3);
				auto hv = v2.getHexValue64();

				mos.writeInt64(hv);
			}
		}

		ptr = e;
	}

	mos.flush();
	return "b64" + mos.getMemoryBlock().toBase64Encoding();
}

}
}
